"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import logging
import os
from copy import deepcopy
from typing import Dict, List, Optional, Union
import torch
import torch.nn as nn
from sapiens_accelerator.utils.imports import (is_4bit_sapiens_available, is_8bit_sapiens_available)
from ..big_modeling import dispatch_model, init_empty_weights
from .dataclasses import BnbQuantizationConfig
from .modeling import (find_tied_parameters, get_balanced_memory, infer_auto_device_map, load_checkpoint_in_model, offload_weight, set_module_tensor_to_device)
logger = logging.getLogger(__name__)
def load_and_quantize_model(model: torch.nn.Module, sapiens_quantization_config: BnbQuantizationConfig, weights_location: Union[str, os.PathLike] = None,
device_map: Optional[Dict[str, Union[int, str, torch.device]]] = None, no_split_module_classes: Optional[List[str]] = None, max_memory: Optional[Dict[Union[int, str], Union[int, str]]] = None,
offload_folder: Optional[Union[str, os.PathLike]] = None, offload_state_dict: bool = False):
    load_in_4bit = sapiens_quantization_config.load_in_4bit
    load_in_8bit = sapiens_quantization_config.load_in_8bit
    if load_in_8bit and not is_8bit_sapiens_available(): raise ImportError("You have a version of `sapiens_machine` that is not compatible with 8bit quantization, make sure you have the latest version of `sapiens_machine` installed.")
    if load_in_4bit and not is_4bit_sapiens_available(): raise ValueError("You have a version of `sapiens_machine` that is not compatible with 4bit quantization, make sure you have the latest version of `sapiens_machine` installed.")
    modules_on_cpu = []
    if isinstance(device_map, dict) and len(device_map.keys()) > 1: modules_on_cpu = [key for key, value in device_map.items() if value in ["disk", "cpu"]]
    if sapiens_quantization_config.skip_modules is None: sapiens_quantization_config.skip_modules = get_keys_to_not_convert(model)
    if load_in_4bit: sapiens_quantization_config.skip_modules.extend(modules_on_cpu)
    modules_to_not_convert = sapiens_quantization_config.skip_modules
    if sapiens_quantization_config.keep_in_fp32_modules is None: sapiens_quantization_config.keep_in_fp32_modules = []
    keep_in_fp32_modules = sapiens_quantization_config.keep_in_fp32_modules
    modules_to_not_convert.extend(keep_in_fp32_modules)
    model.is_loaded_in_4bit = load_in_4bit
    model.is_loaded_in_8bit = load_in_8bit
    model_device = get_parameter_device(model)
    if model_device.type != "meta":
        model = replace_with_sapiens_layers(model, sapiens_quantization_config, modules_to_not_convert=modules_to_not_convert)
        dtype = sapiens_quantization_config.torch_dtype
        for name, param in model.state_dict().items():
            if any(module_to_keep_in_fp32 in name for module_to_keep_in_fp32 in keep_in_fp32_modules):
                param.to(torch.float32)
                if param.dtype != torch.float32:
                    name = name.replace(".weight", "").replace(".bias", "")
                    param = getattr(model, name, None)
                    if param is not None: param.to(torch.float32)
            elif torch.is_floating_point(param): param.to(dtype)
        if model_device.type == "cuda":
            model.cuda(torch.cuda.current_device())
            torch.cuda.empty_cache()
        elif torch.cuda.is_available(): model.to(torch.cuda.current_device())
        else: raise RuntimeError("No GPU found. A GPU is needed for quantization.")
        return model
    elif weights_location is None: raise RuntimeError(f"`weights_location` needs to be the folder path containing the weights of the model, but we found {weights_location} ")
    else:
        with init_empty_weights(): model = replace_with_sapiens_layers(model, sapiens_quantization_config, modules_to_not_convert=modules_to_not_convert)
        device_map = get_quantized_model_device_map(model, sapiens_quantization_config, device_map, max_memory=max_memory, no_split_module_classes=no_split_module_classes)
        if offload_state_dict is None and device_map is not None and "disk" in device_map.values(): offload_state_dict = True
        offload = any(x in list(device_map.values()) for x in ["cpu", "disk"])
        load_checkpoint_in_model(model, weights_location, device_map, dtype=sapiens_quantization_config.torch_dtype, offload_folder=offload_folder,
        offload_state_dict=offload_state_dict, keep_in_fp32_modules=sapiens_quantization_config.keep_in_fp32_modules, offload_8bit_sapiens=load_in_8bit and offload)
        return dispatch_model(model, device_map=device_map, offload_dir=offload_folder)
def get_quantized_model_device_map(model, sapiens_quantization_config, device_map=None, max_memory=None, no_split_module_classes=None):
    if device_map is None:
        if torch.cuda.is_available(): device_map = {"": torch.cuda.current_device()}
        else: raise RuntimeError("No GPU found. A GPU is needed for quantization.")
    if isinstance(device_map, str):
        if device_map not in ["auto", "balanced", "balanced_low_0", "sequential"]: raise ValueError("If passing a string for `device_map`, please choose 'auto', 'balanced', 'balanced_low_0' or 'sequential'.")
        special_dtypes = {}
        special_dtypes.update({name: sapiens_quantization_config.torch_dtype for name, _ in model.named_parameters() if any(m in name for m in sapiens_quantization_config.skip_modules)})
        special_dtypes.update({name: torch.float32 for name, _ in model.named_parameters() if any(m in name for m in sapiens_quantization_config.keep_in_fp32_modules)})
        kwargs = {}
        kwargs["special_dtypes"] = special_dtypes
        kwargs["no_split_module_classes"] = no_split_module_classes
        kwargs["dtype"] = sapiens_quantization_config.target_dtype
        if device_map != "sequential": max_memory = get_balanced_memory(model, low_zero=(device_map == "balanced_low_0"), max_memory=max_memory, **kwargs)
        kwargs["max_memory"] = max_memory
        device_map = infer_auto_device_map(model, **kwargs)
    if isinstance(device_map, dict):
        modules_not_to_convert = sapiens_quantization_config.skip_modules + sapiens_quantization_config.keep_in_fp32_modules
        device_map_without_some_modules = {key: device_map[key] for key in device_map.keys() if key not in modules_not_to_convert}
        for device in ["cpu", "disk"]:
            if device in device_map_without_some_modules.values():
                if sapiens_quantization_config.load_in_4bit: raise
        del device_map_without_some_modules
    return device_map
def replace_with_sapiens_layers(model, sapiens_quantization_config, modules_to_not_convert=None, current_key_name=None):
    if modules_to_not_convert is None: modules_to_not_convert = []
    model, has_been_replaced = _replace_with_sapiens_layers(model, sapiens_quantization_config, modules_to_not_convert, current_key_name)
    return model
def _replace_with_sapiens_layers(model, sapiens_quantization_config, modules_to_not_convert=None, current_key_name=None):
    import sapiens_machine as sapiens
    has_been_replaced = False
    for name, module in model.named_children():
        if current_key_name is None: current_key_name = []
        current_key_name.append(name)
        if isinstance(module, nn.Linear) and name not in modules_to_not_convert:
            current_key_name_str = ".".join(current_key_name)
            proceed = True
            for key in modules_to_not_convert:
                if ((key in current_key_name_str) and (key + "." in current_key_name_str)) or key == current_key_name_str:
                    proceed = False
                    break
            if proceed:
                if sapiens_quantization_config.load_in_8bit: sapiens_module = sapiens.nn.Linear8bitLt(module.in_features, module.out_features, module.bias is not None, has_fp16_weights=False, threshold=sapiens_quantization_config.llm_int8_threshold)
                elif sapiens_quantization_config.load_in_4bit: sapiens_module = sapiens.nn.Linear4bit(module.in_features, module.out_features, module.bias is not None, sapiens_quantization_config.sapiens_4bit_compute_dtype,
                compress_statistics=sapiens_quantization_config.sapiens_4bit_use_double_quant, quant_type=sapiens_quantization_config.sapiens_4bit_quant_type)
                else: raise ValueError("load_in_8bit and load_in_4bit can't be both False")
                sapiens_module.weight.data = module.weight.data
                if module.bias is not None: sapiens_module.bias.data = module.bias.data
                sapiens_module.requires_grad_(False)
                setattr(model, name, sapiens_module)
                has_been_replaced = True
        if len(list(module.children())) > 0:
            _, _has_been_replaced = _replace_with_sapiens_layers(module, sapiens_quantization_config, modules_to_not_convert, current_key_name)
            has_been_replaced = has_been_replaced | _has_been_replaced
        current_key_name.pop(-1)
    return model, has_been_replaced
def get_keys_to_not_convert(model):
    with init_empty_weights(): tied_model = deepcopy(model)
    tied_params = find_tied_parameters(tied_model)
    if isinstance(tied_params, dict): tied_keys = sum(list(tied_params.values()), []) + list(tied_params.keys())
    else: tied_keys = sum(tied_params, [])
    has_tied_params = len(tied_keys) > 0
    is_base_model = False
    if hasattr(model, "base_model_prefix"): is_base_model = not hasattr(model, model.base_model_prefix)
    if (not has_tied_params) and is_base_model: return []
    list_modules = list(model.named_children())
    list_last_module = [list_modules[-1][0]]
    intersection = set(list_last_module) - set(tied_keys)
    list_untouched = list(set(tied_keys)) + list(intersection)
    names_to_remove = [".weight", ".bias"]
    filtered_module_names = []
    for name in list_untouched:
        for name_to_remove in names_to_remove:
            if name_to_remove in name: name = name.replace(name_to_remove, "")
        filtered_module_names.append(name)
    return filtered_module_names
def has_4bit_sapiens_layers(model):
    import sapiens_machine as sapiens
    for m in model.modules():
        if isinstance(m, sapiens.nn.Linear4bit): return True
    return False
def get_parameter_device(parameter: nn.Module): return next(parameter.parameters()).device
def quantize_and_offload_8bit(model, param, param_name, new_dtype, offload_folder, offload_index, fp16_statistics):
    if fp16_statistics is None:
        set_module_tensor_to_device(model, param_name, 0, dtype=new_dtype, value=param)
        tensor_name = param_name
        module = model
        if "." in tensor_name:
            splits = tensor_name.split(".")
            for split in splits[:-1]:
                new_module = getattr(module, split)
                if new_module is None: raise ValueError(f"{module} has no attribute {split}.")
                module = new_module
            tensor_name = splits[-1]
        module._parameters[tensor_name].requires_grad = False
        offload_weight(module._parameters[tensor_name], param_name, offload_folder, index=offload_index)
        if hasattr(module._parameters[tensor_name], "SCB"): offload_weight(module._parameters[tensor_name].SCB, param_name.replace("weight", "SCB"), offload_folder, index=offload_index)
    else:
        offload_weight(param, param_name, offload_folder, index=offload_index)
        offload_weight(fp16_statistics, param_name.replace("weight", "SCB"), offload_folder, index=offload_index)
    set_module_tensor_to_device(model, param_name, "meta", dtype=new_dtype, value=torch.empty(*param.size()))
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from pathlib import Path
from .config_args import default_config_file, load_config_from_file
from .config_utils import SubcommandHelpFormatter
description = "Update an existing config file with the latest defaults while maintaining the old configuration."
def update_config(args):
    config_file = args.config_file
    if config_file is None and Path(default_config_file).exists(): config_file = default_config_file
    elif not Path(config_file).exists(): raise ValueError(f"The passed config file located at {config_file} doesn't exist.")
    config = load_config_from_file(config_file)
    if config_file.endswith(".json"): config.to_json_file(config_file)
    else: config.to_yaml_file(config_file)
    return config_file
def update_command_parser(parser, parents):
    parser = parser.add_parser("update", parents=parents, help=description, formatter_class=SubcommandHelpFormatter)
    parser.add_argument("--config_file", default=None, help=("The path to the config file to update. Will default to a file named default_config.yaml in the cache location, which is the content of the environment `HF_HOME` suffixed with 'sapiens_accelerator', or if you don't have such an environment variable, your cache directory ('~/.cache' or the content of `XDG_CACHE_HOME`) suffixed with 'huggingface'."))
    parser.set_defaults(func=update_config_command)
    return parser
def update_config_command(args):
    config_file = update_config(args)
    print(f"Sucessfully updated the configuration file at {config_file}.")
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import numpy as np
import torch
from torch.utils.data import DataLoader
from sapiens_accelerator.utils.dataclasses import DistributedType
class RegressionDataset:
    def __init__(self, a=2, b=3, length=64, seed=None):
        rng = np.random.default_rng(seed)
        self.length = length
        self.x = rng.normal(size=(length,)).astype(np.float32)
        self.y = a * self.x + b + rng.normal(scale=0.1, size=(length,)).astype(np.float32)
    def __len__(self): return self.length
    def __getitem__(self, i): return {"x": self.x[i], "y": self.y[i]}
class RegressionModel4XPU(torch.nn.Module):
    def __init__(self, a=0, b=0, double_output=False):
        super().__init__()
        self.a = torch.nn.Parameter(torch.tensor([2, 3]).float())
        self.b = torch.nn.Parameter(torch.tensor([2, 3]).float())
        self.first_batch = True
    def forward(self, x=None):
        if self.first_batch: self.first_batch = False
        return x * self.a[0] + self.b[0]
class RegressionModel(torch.nn.Module):
    def __init__(self, a=0, b=0, double_output=False):
        super().__init__()
        self.a = torch.nn.Parameter(torch.tensor(a).float())
        self.b = torch.nn.Parameter(torch.tensor(b).float())
        self.first_batch = True
    def forward(self, x=None):
        if self.first_batch: self.first_batch = False
        return x * self.a + self.b
def mocked_dataloaders(accelerator, batch_size: int = 16):
    from datasets import load_dataset
    from transformers import AutoTokenizer
    tokenizer = AutoTokenizer.from_pretrained("bert-base-cased")
    data_files = {"train": "tests/test_samples/MRPC/train.csv", "validation": "tests/test_samples/MRPC/dev.csv"}
    datasets = load_dataset("csv", data_files=data_files)
    label_list = datasets["train"].unique("label")
    label_to_id = {v: i for i, v in enumerate(label_list)}
    def tokenize_function(examples):
        outputs = tokenizer(examples["sentence1"], examples["sentence2"], truncation=True, max_length=None, padding="max_length")
        if "label" in examples: outputs["labels"] = [label_to_id[l] for l in examples["label"]]
        return outputs
    tokenized_datasets = datasets.map(tokenize_function, batched=True, remove_columns=["sentence1", "sentence2", "label"])
    def collate_fn(examples):
        if accelerator.distributed_type == DistributedType.XLA: return tokenizer.pad(examples, padding="max_length", max_length=128, return_tensors="pt")
        return tokenizer.pad(examples, padding="longest", return_tensors="pt")
    train_dataloader = DataLoader(tokenized_datasets["train"], shuffle=True, collate_fn=collate_fn, batch_size=2)
    eval_dataloader = DataLoader(tokenized_datasets["validation"], shuffle=False, collate_fn=collate_fn, batch_size=1)
    return train_dataloader, eval_dataloader
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import asyncio
import inspect
import io
import os
import shutil
import subprocess
import sys
import tempfile
import unittest
from contextlib import contextmanager
from functools import partial
from pathlib import Path
from typing import List, Union
from unittest import mock
import torch
import sapiens_accelerator
from ..state import AcceleratorState, PartialState
from ..utils import (gather, is_sapiens_available, is_clearml_available, is_comet_ml_available, is_cuda_available, is_datasets_available, is_deepspeed_available,
is_dvclive_available, is_import_timer_available, is_mlu_available, is_mps_available, is_musa_available, is_npu_available, is_pandas_available, is_pippy_available,
is_schedulefree_available, is_tensorboard_available, is_timm_available, is_torch_version, is_torch_xla_available, is_torchdata_stateful_dataloader_available,
is_torchvision_available, is_transformer_engine_available, is_transformers_available, is_triton_available, is_wandb_available, is_xpu_available, str_to_bool)
def get_backend():
    if is_torch_xla_available(): return "xla", torch.cuda.device_count(), torch.cuda.memory_allocated
    elif is_cuda_available(): return "cuda", torch.cuda.device_count(), torch.cuda.memory_allocated
    elif is_mps_available(min_version="2.0"): return "mps", 1, torch.mps.current_allocated_memory
    elif is_mps_available(): return "mps", 1, lambda: 0
    elif is_mlu_available(): return "mlu", torch.mlu.device_count(), torch.mlu.memory_allocated
    elif is_musa_available(): return "musa", torch.musa.device_count(), torch.musa.memory_allocated
    elif is_npu_available(): return "npu", torch.npu.device_count(), torch.npu.memory_allocated
    elif is_xpu_available(): return "xpu", torch.xpu.device_count(), torch.xpu.memory_allocated
    else: return "cpu", 1, lambda: 0
torch_device, device_count, memory_allocated_func = get_backend()
def get_launch_command(**kwargs) -> list:
    command = ["sapiens_accelerator", "launch"]
    for k, v in kwargs.items():
        if isinstance(v, bool) and v: command.append(f"--{k}")
        elif v is not None: command.append(f"--{k}={v}")
    return command
DEFAULT_LAUNCH_COMMAND = get_launch_command(num_processes=device_count, monitor_interval=0.1)
def parse_flag_from_env(key, default=False):
    try: value = os.environ[key]
    except KeyError: _value = default
    else:
        try: _value = str_to_bool(value)
        except ValueError: raise ValueError(f"If set, {key} must be yes or no.")
    return _value
_run_slow_tests = parse_flag_from_env("RUN_SLOW", default=False)
def skip(test_case): return unittest.skip("Test was skipped")(test_case)
def slow(test_case): return unittest.skipUnless(_run_slow_tests, "test is slow")(test_case)
def require_cpu(test_case): return unittest.skipUnless(torch_device == "cpu", "test requires only a CPU")(test_case)
def require_non_cpu(test_case): return unittest.skipUnless(torch_device != "cpu", "test requires a GPU")(test_case)
def require_cuda(test_case): return unittest.skipUnless(is_cuda_available() and not is_torch_xla_available(), "test requires a GPU")(test_case)
def require_xpu(test_case): return unittest.skipUnless(is_xpu_available(), "test requires a XPU")(test_case)
def require_non_xpu(test_case): return unittest.skipUnless(torch_device != "xpu", "test requires a non-XPU")(test_case)
def require_mlu(test_case): return unittest.skipUnless(is_mlu_available(), "test require a MLU")(test_case)
def require_musa(test_case): return unittest.skipUnless(is_musa_available(), "test require a MUSA")(test_case)
def require_npu(test_case): return unittest.skipUnless(is_npu_available(), "test require a NPU")(test_case)
def require_mps(test_case): return unittest.skipUnless(is_mps_available(), "test requires a `mps` backend support in `torch`")(test_case)
def require_huggingface_suite(test_case): return unittest.skipUnless(is_transformers_available() and is_datasets_available(), "test requires the HF suite")(test_case)
def require_transformers(test_case): return unittest.skipUnless(is_transformers_available(), "test requires the transformers library")(test_case)
def require_timm(test_case): return unittest.skipUnless(is_timm_available(), "test requires the timm library")(test_case)
def require_torchvision(test_case): return unittest.skipUnless(is_torchvision_available(), "test requires the torchvision library")(test_case)
def require_triton(test_case): return unittest.skipUnless(is_triton_available(), "test requires the triton library")(test_case)
def require_schedulefree(test_case): return unittest.skipUnless(is_schedulefree_available(), "test requires the schedulefree library")(test_case)
def require_sapiens(test_case): return unittest.skipUnless(is_sapiens_available(), "test requires the sapiens_machine library")(test_case)
def require_tpu(test_case): return unittest.skipUnless(is_torch_xla_available(check_is_tpu=True), "test requires TPU")(test_case)
def require_non_torch_xla(test_case): return unittest.skipUnless(not is_torch_xla_available(), "test requires an env without TorchXLA")(test_case)
def require_single_device(test_case): return unittest.skipUnless(torch_device != "cpu" and device_count == 1, "test requires a hardware accelerator")(test_case)
def require_single_gpu(test_case): return unittest.skipUnless(torch.cuda.device_count() == 1, "test requires a GPU")(test_case)
def require_single_xpu(test_case): return unittest.skipUnless(torch.xpu.device_count() == 1, "test requires a XPU")(test_case)
def require_multi_device(test_case): return unittest.skipUnless(device_count > 1, "test requires multiple hardware accelerators")(test_case)
def require_multi_gpu(test_case): return unittest.skipUnless(torch.cuda.device_count() > 1, "test requires multiple GPUs")(test_case)
def require_multi_xpu(test_case): return unittest.skipUnless(torch.xpu.device_count() > 1, "test requires multiple XPUs")(test_case)
def require_deepspeed(test_case): return unittest.skipUnless(is_deepspeed_available(), "test requires DeepSpeed")(test_case)
def require_fsdp(test_case): return unittest.skipUnless(is_torch_version(">=", "1.12.0"), "test requires torch version >= 1.12.0")(test_case)
def require_torch_min_version(test_case=None, version=None):
    if test_case is None: return partial(require_torch_min_version, version=version)
    return unittest.skipUnless(is_torch_version(">=", version), f"test requires torch version >= {version}")(test_case)
def require_tensorboard(test_case): return unittest.skipUnless(is_tensorboard_available(), "test requires Tensorboard")(test_case)
def require_wandb(test_case): return unittest.skipUnless(is_wandb_available(), "test requires wandb")(test_case)
def require_comet_ml(test_case): return unittest.skipUnless(is_comet_ml_available(), "test requires comet_ml")(test_case)
def require_clearml(test_case): return unittest.skipUnless(is_clearml_available(), "test requires clearml")(test_case)
def require_dvclive(test_case): return unittest.skipUnless(is_dvclive_available(), "test requires dvclive")(test_case)
def require_pandas(test_case): return unittest.skipUnless(is_pandas_available(), "test requires pandas")(test_case)
def require_pippy(test_case): return unittest.skipUnless(is_pippy_available(), "test requires pippy")(test_case)
def require_import_timer(test_case): return unittest.skipUnless(is_import_timer_available(), "test requires tuna interpreter")(test_case)
def require_transformer_engine(test_case): return unittest.skipUnless(is_transformer_engine_available(), "test requires transformers engine")(test_case)
_atleast_one_tracker_available = (any([is_wandb_available(), is_tensorboard_available()]) and not is_comet_ml_available())
def require_trackers(test_case): return unittest.skipUnless(_atleast_one_tracker_available, "test requires at least one tracker to be available and for `comet_ml` to not be installed")(test_case)
def require_torchdata_stateful_dataloader(test_case): return unittest.skipUnless(is_torchdata_stateful_dataloader_available(), "test requires torchdata.stateful_dataloader")(test_case)
class TempDirTestCase(unittest.TestCase):
    clear_on_setup = True
    @classmethod
    def setUpClass(cls): cls.tmpdir = Path(tempfile.mkdtemp())
    @classmethod
    def tearDownClass(cls):
        if os.path.exists(cls.tmpdir): shutil.rmtree(cls.tmpdir)
    def setUp(self):
        if self.clear_on_setup:
            for path in self.tmpdir.glob("**/*"):
                if path.is_file(): path.unlink()
                elif path.is_dir(): shutil.rmtree(path)
class SapiensAcceleratorTestCase(unittest.TestCase):
    def tearDown(self):
        super().tearDown()
        AcceleratorState._reset_state()
        PartialState._reset_state()
class MockingTestCase(unittest.TestCase):
    def add_mocks(self, mocks: Union[mock.Mock, List[mock.Mock]]):
        self.mocks = mocks if isinstance(mocks, (tuple, list)) else [mocks]
        for m in self.mocks:
            m.start()
            self.addCleanup(m.stop)
def are_the_same_tensors(tensor):
    state = AcceleratorState()
    tensor = tensor[None].clone().to(state.device)
    tensors = gather(tensor).cpu()
    tensor = tensor[0].cpu()
    for i in range(tensors.shape[0]):
        if not torch.equal(tensors[i], tensor): return False
    return True
class _RunOutput:
    def __init__(self, returncode, stdout, stderr):
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
async def _read_stream(stream, callback):
    while True:
        line = await stream.readline()
        if line: callback(line)
        else: break
async def _stream_subprocess(cmd, env=None, stdin=None, timeout=None, quiet=False, echo=False) -> _RunOutput:
    if echo: print("\nRunning: ", " ".join(cmd))
    p = await asyncio.create_subprocess_exec(cmd[0], *cmd[1:], stdin=stdin, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE, env=env)
    out = []
    err = []
    def tee(line, sink, pipe, label=""):
        line = line.decode("utf-8").rstrip()
        sink.append(line)
        if not quiet: print(label, line, file=pipe)
    await asyncio.wait([asyncio.create_task(_read_stream(p.stdout, lambda l: tee(l, out, sys.stdout, label="stdout:"))), asyncio.create_task(_read_stream(p.stderr, lambda l: tee(l, err, sys.stderr, label="stderr:")))], timeout=timeout)
    return _RunOutput(await p.wait(), out, err)
def execute_subprocess_async(cmd: list, env=None, stdin=None, timeout=180, quiet=False, echo=True) -> _RunOutput:
    for i, c in enumerate(cmd):
        if isinstance(c, Path): cmd[i] = str(c)
    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(_stream_subprocess(cmd, env=env, stdin=stdin, timeout=timeout, quiet=quiet, echo=echo))
    cmd_str = " ".join(cmd)
    if result.returncode > 0:
        stderr = "\n".join(result.stderr)
        raise RuntimeError(f"'{cmd_str}' failed with returncode {result.returncode}\n\nThe combined stderr from workers follows:\n{stderr}")
    return result
class SubprocessCallException(Exception): pass
def run_command(command: List[str], return_stdout=False, env=None):
    for i, c in enumerate(command):
        if isinstance(c, Path): command[i] = str(c)
    if env is None: env = os.environ.copy()
    try:
        output = subprocess.check_output(command, stderr=subprocess.STDOUT, env=env)
        if return_stdout:
            if hasattr(output, "decode"): output = output.decode("utf-8")
            return output
    except subprocess.CalledProcessError as e: raise SubprocessCallException(f"Command `{' '.join(command)}` failed with the following error:\n\n{e.output.decode()}") from e
def path_in_sapiens_accelerator_package(*components: str) -> Path:
    sapiens_accelerator_package_dir = Path(inspect.getfile(sapiens_accelerator)).parent
    return sapiens_accelerator_package_dir.joinpath(*components)
@contextmanager
def assert_exception(exception_class: Exception, msg: str = None) -> bool:
    was_ran = False
    try:
        yield
        was_ran = True
    except Exception as e:
        assert isinstance(e, exception_class), f"Expected exception of type {exception_class} but got {type(e)}"
        if msg is not None: assert msg in str(e), f"Expected message '{msg}' to be in exception but got '{str(e)}'"
    if was_ran: raise AssertionError(f"Expected exception of type {exception_class} but ran without issue.")
def capture_call_output(func, *args, **kwargs):
    captured_output = io.StringIO()
    original_stdout = sys.stdout
    try:
        sys.stdout = captured_output
        func(*args, **kwargs)
    except Exception as e: raise e
    finally: sys.stdout = original_stdout
    return captured_output.getvalue()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from typing import List
from .keymap import KEYMAP, get_character
def mark(key: str):
    def decorator(func):
        handle = getattr(func, "handle_key", [])
        handle += [key]
        func.handle_key = handle
        return func
    return decorator
def mark_multiple(*keys: List[str]):
    def decorator(func):
        handle = getattr(func, "handle_key", [])
        handle += keys
        func.handle_key = handle
        return func
    return decorator
class KeyHandler(type):
    def __new__(cls, name, bases, attrs):
        new_cls = super().__new__(cls, name, bases, attrs)
        if not hasattr(new_cls, "key_handler"): new_cls.key_handler = {}
        new_cls.handle_input = KeyHandler.handle_input
        for value in attrs.values():
            handled_keys = getattr(value, "handle_key", [])
            for key in handled_keys: new_cls.key_handler[key] = value
        return new_cls
    @staticmethod
    def handle_input(cls):
        char = get_character()
        if char != KEYMAP["undefined"]: char = ord(char)
        handler = cls.key_handler.get(char)
        if handler:
            cls.current_selection = char
            return handler(cls)
        else: return None
def register(cls): return KeyHandler(cls.__name__, cls.__bases__, cls.__dict__.copy())
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from __future__ import annotations
import contextlib
import functools
import json
import math
import os
import re
import shutil
import sys
import warnings
from collections import OrderedDict
from contextlib import contextmanager
from functools import partial
from types import MethodType
from typing import Any, Callable, Union
import torch
import torch.utils.hooks as hooks
from huggingface_hub import split_torch_state_dict_into_shards
from packaging import version
from .checkpointing import load_accelerator_state, load_custom_state, save_accelerator_state, save_custom_state
from .data_loader import DataLoaderDispatcher, prepare_data_loader, skip_first_batches
from .hooks import AlignDevicesHook
from .logging import get_logger
from .optimizer import SapiensAcceleratordOptimizer
from .scheduler import SapiensAcceleratordScheduler
from .state import AcceleratorState, GradientState, PartialState
from .tracking import LOGGER_TYPE_TO_CLASS, GeneralTracker, filter_trackers
from .utils import (MODEL_NAME, SAFE_WEIGHTS_INDEX_NAME, SAFE_WEIGHTS_NAME, SAFE_WEIGHTS_PATTERN_NAME, WEIGHTS_INDEX_NAME, WEIGHTS_NAME, WEIGHTS_PATTERN_NAME,
AutocastKwargs, DataLoaderConfiguration, DeepSpeedPlugin, DistributedDataParallelKwargs, DistributedType, DynamoBackend, FP8RecipeKwargs, FullyShardedDataParallelPlugin,
GradientAccumulationPlugin, GradScalerKwargs, InitProcessGroupKwargs, KwargsHandler, LoggerType, MegatronLMPlugin, PrecisionType, ProfileKwargs, ProjectConfiguration,
RNGType, TorchDynamoPlugin, apply_fp8_autowrap, check_os_kernel, clean_state_dict_for_safetensors, compare_versions, convert_model, convert_outputs_to_fp32,
extract_model_from_parallel, gather, gather_object, get_mixed_precision_context_manager, get_pretty_name, is_bf16_available, is_sapiens_machine_multi_backend_available,
is_deepspeed_available, is_ipex_available, is_lomo_available, is_megatron_lm_available, is_mlu_available, is_msamp_available, is_musa_available, is_npu_available,
is_torch_version, is_torch_xla_available, is_transformer_engine_available, is_xpu_available, load_fsdp_model, load_fsdp_optimizer, pad_across_processes,
parse_choice_from_env, recursively_apply, reduce, release_memory, save, save_fsdp_model, save_fsdp_optimizer, wait_for_everyone)
from .utils.constants import FSDP_PYTORCH_VERSION, PROFILE_PATTERN_NAME
from .utils.modeling import get_state_dict_offloaded_model
from .utils.other import is_compiled_module
if is_deepspeed_available(): from .utils import (DeepSpeedEngineWrapper, DeepSpeedOptimizerWrapper, DeepSpeedSchedulerWrapper, DummyOptim, DummyScheduler)
if is_megatron_lm_available(): from .utils import (MegatronEngine, MegatronLMDummyDataLoader, MegatronLMDummyScheduler, MegatronLMOptimizerWrapper, MegatronLMSchedulerWrapper,
megatron_lm_initialize, megatron_lm_prepare_data_loader, megatron_lm_prepare_model_optimizer_scheduler)
from torch.distributed.algorithms.join import Join
if is_torch_xla_available():
    import torch_xla.amp as xamp
    import torch_xla.core.xla_model as xm
    import torch_xla.distributed.xla_multiprocessing as xmp
if is_npu_available(check_device=False): import torch_npu
try: from torch.optim.lr_scheduler import LRScheduler
except ImportError: from torch.optim.lr_scheduler import _LRScheduler as LRScheduler
logger = get_logger(__name__)
_split_batches = object()
_dispatch_batches = object()
_even_batches = object()
_use_seedable_sampler = object()
class Accelerator:
    def __init__(self, device_placement: bool = True, split_batches: bool = _split_batches, mixed_precision: PrecisionType | str | None = None, gradient_accumulation_steps: int = 1,
    cpu: bool = False, dataloader_config: DataLoaderConfiguration | None = None, deepspeed_plugin: DeepSpeedPlugin | dict[str, DeepSpeedPlugin] | None = None,
    fsdp_plugin: FullyShardedDataParallelPlugin | None = None, megatron_lm_plugin: MegatronLMPlugin | None = None, rng_types: list[str | RNGType] | None = None,
    log_with: str | LoggerType | GeneralTracker | list[str | LoggerType | GeneralTracker] | None = None, project_dir: str | os.PathLike | None = None,
    project_config: ProjectConfiguration | None = None, gradient_accumulation_plugin: GradientAccumulationPlugin | None = None, step_scheduler_with_optimizer: bool = True,
    kwargs_handlers: list[KwargsHandler] | None = None, dynamo_backend: DynamoBackend | str | None = None, deepspeed_plugins: DeepSpeedPlugin | dict[str, DeepSpeedPlugin] | None = None):
        self.trackers = []
        if project_config is not None: self.project_configuration = project_config
        else: self.project_configuration = ProjectConfiguration(project_dir=project_dir)
        if project_dir is not None and self.project_dir is None: self.project_configuration.set_directories(project_dir)
        if mixed_precision is not None:
            mixed_precision = str(mixed_precision)
            if mixed_precision not in PrecisionType: raise ValueError(f"Unknown mixed_precision mode: {mixed_precision}. Choose between {PrecisionType.list()}")
        dynamo_plugin = TorchDynamoPlugin() if dynamo_backend is None else TorchDynamoPlugin(backend=dynamo_backend)
        if deepspeed_plugins is not None and deepspeed_plugin is not None: raise ValueError("You cannot pass in both `deepspeed_plugins` and `deepspeed_plugin`.")
        elif deepspeed_plugin is not None: deepspeed_plugins = deepspeed_plugin
        if deepspeed_plugins is None:
            if PartialState._shared_state != {} and PartialState().distributed_type == DistributedType.DEEPSPEED: deepspeed_plugins = AcceleratorState().deepspeed_plugins
            else: deepspeed_plugins = (DeepSpeedPlugin() if os.environ.get("SAPIENS_ACCELERATOR_USE_DEEPSPEED", "false") == "true" else None)
        else:
            if (PartialState().distributed_type == DistributedType.DEEPSPEED and AcceleratorState._shared_state != {} and AcceleratorState().deepspeed_plugins is not None): raise NotImplementedError("You cannot pass in a `deepspeed_plugin` when creating a second `Accelerator`. Please make sure the first `Accelerator` is initialized with all the plugins you want to use.")
            if isinstance(deepspeed_plugins, dict):
                for plugin in deepspeed_plugins.values():
                    if not isinstance(plugin, DeepSpeedPlugin): raise TypeError("`deepspeed_plugin` must be a DeepSpeedPlugin object.")
        if deepspeed_plugins is not None:
            os.environ["SAPIENS_ACCELERATOR_USE_DEEPSPEED"] = "true"
            if not is_deepspeed_available(): raise ImportError("DeepSpeed is not installed => run `pip install deepspeed` or build it from source.")
            if is_mlu_available():
                if compare_versions("deepspeed-mlu", "<", "0.10.1"): raise ImportError("DeepSpeed MLU version must be >= 0.10.1. Please update DeepSpeed MLU.")
            elif is_musa_available():
                if compare_versions("deepspeed", ">", "0.14.3"): raise ImportError("DeepSpeed MUSA version must be <= 0.14.3. Please downgrade DeepSpeed.")
            elif compare_versions("deepspeed", "<", "0.9.3"): raise ImportError("DeepSpeed version must be >= 0.9.3. Please update DeepSpeed.")
            mixed_precision = (os.environ.get("SAPIENS_ACCELERATOR_MIXED_PRECISION", "no") if mixed_precision is None else mixed_precision)
            if not isinstance(deepspeed_plugins, dict):
                deepspeed_plugins.set_mixed_precision(mixed_precision)
                deepspeed_plugins.select(_from_accelerator_state=True)
            else:
                for plugin in deepspeed_plugins.values(): plugin.set_mixed_precision(mixed_precision)
                first_plugin = next(iter(deepspeed_plugins.values()))
                first_plugin.select(_from_accelerator_state=True)
            self.deepspeed_engine_wrapped = None
        if os.environ.get("SAPIENS_ACCELERATOR_USE_FSDP", "false") == "true" or isinstance(fsdp_plugin, FullyShardedDataParallelPlugin):
            if not is_torch_version(">=", FSDP_PYTORCH_VERSION): raise ValueError(f"FSDP requires PyTorch >= {FSDP_PYTORCH_VERSION}")
        if fsdp_plugin is None: fsdp_plugin = (FullyShardedDataParallelPlugin() if os.environ.get("SAPIENS_ACCELERATOR_USE_FSDP", "false") == "true" else None)
        else:
            if not isinstance(fsdp_plugin, FullyShardedDataParallelPlugin): raise TypeError("`fsdp_plugin` must be a FullyShardedDataParallelPlugin object.")
            os.environ["SAPIENS_ACCELERATOR_USE_FSDP"] = "true"
        if megatron_lm_plugin is None: megatron_lm_plugin = (MegatronLMPlugin() if os.environ.get("SAPIENS_ACCELERATOR_USE_MEGATRON_LM", "false") == "true" else None)
        else:
            if not isinstance(megatron_lm_plugin, MegatronLMPlugin): raise TypeError("`megatron_lm_plugin` must be a MegatronLMPlugin object.")
            os.environ["SAPIENS_ACCELERATOR_USE_MEGATRON_LM"] = "true"
        if megatron_lm_plugin:
            if not is_megatron_lm_available(): raise ImportError("Megatron is not installed. please build it from source.")
        self.ddp_handler = None
        self.scaler_handler = None
        self.init_handler = None
        self.fp8_recipe_handler = None
        self.autocast_handler = None
        self.profile_handler = None
        self.has_lomo_optimizer = False
        if kwargs_handlers is not None:
            for handler in kwargs_handlers:
                assert isinstance(handler, KwargsHandler), f"Unsupported kwargs handler passed: {handler}, must be one that inherits `sapiens_accelerator.utils.KwargsHandler`."
                if isinstance(handler, DistributedDataParallelKwargs):
                    if self.ddp_handler is not None: raise ValueError("You can only pass one `DistributedDataParallelKwargs` in `kwargs_handler`.")
                    else: self.ddp_handler = handler
                elif isinstance(handler, GradScalerKwargs):
                    if self.scaler_handler is not None: raise ValueError("You can only pass one `GradScalerKwargs` in `kwargs_handler`.")
                    else: self.scaler_handler = handler
                elif isinstance(handler, InitProcessGroupKwargs):
                    if self.init_handler is not None: raise ValueError("You can only pass one `InitProcessGroupKwargs` in `kwargs_handler`.")
                    else: self.init_handler = handler
                elif isinstance(handler, FP8RecipeKwargs):
                    if self.fp8_recipe_handler is not None: raise ValueError("You can only pass one `FP8RecipeKwargs` in `kwargs_handler`.")
                    else: self.fp8_recipe_handler = handler
                elif isinstance(handler, AutocastKwargs):
                    if self.autocast_handler is not None: raise ValueError("You can only pass one `AutocastKwargs` in `kwargs_handler`.")
                    else: self.autocast_handler = handler
                elif isinstance(handler, ProfileKwargs):
                    if self.profile_handler is not None: raise ValueError("You can only pass one `ProfileKwargs` in `kwargs_handler`.")
                    else: self.profile_handler = handler
        kwargs = self.init_handler.to_kwargs() if self.init_handler is not None else {}
        self.state = AcceleratorState(mixed_precision=mixed_precision, cpu=cpu, dynamo_plugin=dynamo_plugin, deepspeed_plugin=deepspeed_plugins,
        fsdp_plugin=fsdp_plugin, megatron_lm_plugin=megatron_lm_plugin, _from_accelerator=True, **kwargs)
        if self.state.mixed_precision == "fp8" and self.fp8_recipe_handler is None: self.fp8_recipe_handler = FP8RecipeKwargs()
        self.delayed_fp8_autocast = False
        if self.fp8_recipe_handler is not None:
            if self.state.mixed_precision != "fp8" and (self.distributed_type not in (DistributedType.FSDP, DistributedType.DEEPSPEED)): raise ValueError("Passing in a `FP8RecipeKwargs` object requires setting `mixed_precision='fp8'`.")
            self.delayed_fp8_autocast = self.fp8_recipe_handler.backend == "TE" and self.distributed_type in (DistributedType.MULTI_GPU, DistributedType.FSDP)
        trackers = filter_trackers(log_with, self.logging_dir)
        self.log_with = trackers
        if ((mixed_precision != "bf16") and getattr(self.state, "downcast_bfloat", False) and (self.state.distributedType != DistributedType.XLA)): raise ValueError("Can only use `downcast_bf16` when using `mixed_precision='bf16'` and on a TPU")
        if gradient_accumulation_plugin is not None:
            if gradient_accumulation_steps != 1: raise ValueError("You can only pass one of `gradient_accumulation_steps` and `gradient_accumulation_plugin`. Please only pass in the created `GradientAccumulationPlugin` object.")
        else:
            gradient_accumulation_steps = int(parse_choice_from_env("SAPIENS_ACCELERATOR_GRADIENT_ACCUMULATION_STEPS", gradient_accumulation_steps))
            gradient_accumulation_plugin = GradientAccumulationPlugin(num_steps=gradient_accumulation_steps)
        self.gradient_state = GradientState(gradient_accumulation_plugin=gradient_accumulation_plugin)
        self.device_placement = device_placement
        if dataloader_config is None: dataloader_config = DataLoaderConfiguration()
        self.dataloader_config = dataloader_config
        self.step_scheduler_with_optimizer = step_scheduler_with_optimizer
        self.scaler = None
        self.native_amp = False
        if (self.state.mixed_precision == "fp16" and self.device.type != "cpu" and self.distributed_type not in (DistributedType.DEEPSPEED, DistributedType.MEGATRON_LM)):
            self.native_amp = True
            if self.device.type not in ("xpu", "cuda", "npu", "xla", "mlu", "musa") or is_torch_xla_available(check_is_tpu=True): raise ValueError(f"fp16 mixed precision requires a GPU (not {self.device.type!r}).")
            kwargs = self.scaler_handler.to_kwargs() if self.scaler_handler is not None else {}
            if self.distributed_type == DistributedType.FSDP:
                from torch.distributed.fsdp.sharded_grad_scaler import ShardedGradScaler
                self.scaler = ShardedGradScaler(**kwargs)
            elif is_torch_xla_available(check_is_gpu=True): self.scaler = xamp.GradScaler(**kwargs)
            elif is_mlu_available(): self.scaler = torch.mlu.amp.GradScaler(**kwargs)
            elif is_musa_available(): self.scaler = torch.musa.amp.GradScaler(**kwargs)
            elif is_npu_available(): self.scaler = torch.npu.amp.GradScaler(**kwargs)
            elif is_xpu_available(): self.scaler = torch.amp.GradScaler("xpu", **kwargs)
            else:
                if version.parse(torch.__version__) > version.parse("2.3"): self.scaler = torch.amp.GradScaler("cuda", **kwargs)
                else: self.scaler = torch.cuda.amp.GradScaler(**kwargs)
        elif self.state.mixed_precision == "bf16" and self.distributed_type not in (DistributedType.DEEPSPEED, DistributedType.MEGATRON_LM):
            if self.device.type in ["cpu", "xpu"]: self.native_amp = True
            else: self.native_amp = is_bf16_available(True)
            if mixed_precision == "bf16" and not self.native_amp and not is_torch_xla_available(): raise ValueError("bf16 mixed precision requires PyTorch >= 1.10 and a supported device.")
        elif self.state.mixed_precision == "fp8":
            self.native_amp = True
            if self.fp8_backend == "MSAMP":
                if self.distributed_type == DistributedType.FSDP: raise NotImplementedError("`sapiens_accelerator` + `MS-AMP` + `FSDP` is not supported at this time. Please consider using deepspeed, which is supported.")
                elif self.distributed_type != DistributedType.DEEPSPEED:
                    if version.parse(torch.__version__) > version.parse("2.3"): self.scaler = torch.amp.GradScaler("cuda")
                    else: self.scaler = torch.cuda.amp.GradScaler()
        self.step = 0
        self._optimizers = []
        self._models = []
        self._schedulers = []
        self._dataloaders = []
        self._custom_objects = []
        self._load_model_state_pre_hook = OrderedDict()
        self._save_model_state_pre_hook = OrderedDict()
        self.rng_types = rng_types
        if self.rng_types is None: self.rng_types = ["generator"]
        self.flag_tensor = None
        check_os_kernel()
    @property
    def deepspeed_plugin(self): return self.state.deepspeed_plugin
    @property
    def use_distributed(self): return self.state.use_distributed
    @property
    def distributed_type(self): return self.state.distributed_type
    @property
    def num_processes(self): return self.state.num_processes
    @property
    def process_index(self): return self.state.process_index
    @property
    def local_process_index(self): return self.state.local_process_index
    @property
    def device(self): return self.state.device
    @property
    def split_batches(self): return self.dataloader_config.split_batches
    @property
    def dispatch_batches(self): return self.dataloader_config.dispatch_batches
    @property
    def even_batches(self): return self.dataloader_config.even_batches
    @even_batches.setter
    def even_batches(self, value: bool): self.dataloader_config.even_batches = value
    @property
    def use_seedable_sampler(self): return self.dataloader_config.use_seedable_sampler
    @property
    def non_blocking(self): return self.dataloader_config.non_blocking
    @property
    def use_stateful_dataloader(self):
        if hasattr(self.dataloader_config, "use_stateful_dataloader"): return self.dataloader_config.use_stateful_dataloader
        return False
    @property
    def project_dir(self): return self.project_configuration.project_dir
    @property
    def logging_dir(self): return self.project_configuration.logging_dir
    @property
    def save_iteration(self): return self.project_configuration.iteration
    @property
    def is_main_process(self): return self.state.is_main_process
    @property
    def is_local_main_process(self): return self.state.is_local_main_process
    @property
    def is_last_process(self): return self.process_index == self.num_processes - 1
    @property
    def mixed_precision(self): return self.state.mixed_precision
    @contextmanager
    def split_between_processes(self, inputs: list | tuple | dict | torch.Tensor, apply_padding: bool = False):
        with PartialState().split_between_processes(inputs, apply_padding=apply_padding) as inputs: yield inputs
    def on_main_process(self, function: Callable[..., Any] = None):
        if function is None:
            if "Accelerator." in self.__qualname__: function = self
            else: raise ValueError("The `on_main_process` decorator must be called with a function on an instantiated `Accelerator` object.")
        def _inner(*args, **kwargs): return PartialState().on_main_process(function)(*args, **kwargs)
        return _inner
    def on_local_main_process(self, function: Callable[..., Any] = None):
        if function is None:
            if "Accelerator." in self.__qualname__: function = self
            else: raise ValueError("The `on_local_main_process` decorator must be called with a function on an instantiated `Accelerator` object.")
        def _inner(*args, **kwargs): return PartialState().on_local_main_process(function)(*args, **kwargs)
        return _inner
    def on_last_process(self, function: Callable[..., Any]):
        if function is None:
            if "Accelerator." in self.__qualname__: function = self
            else: raise ValueError("The `on_last_process` decorator must be called with a function on an instantiated `Accelerator` object.")
        def _inner(*args, **kwargs): return PartialState().on_last_process(function)(*args, **kwargs)
        return _inner
    def on_process(self, function: Callable[..., Any] = None, process_index: int = None):
        if (self is not None) and (process_index is not None) and (function is None): return partial(self.on_process, process_index=process_index)
        if function is None:
            if "Accelerator." in self.__qualname__: function = self
            else: raise ValueError("The `on_main_process` decorator must be called with a function on an instantiated `Accelerator` object.")
        def _inner(*args, **kwargs): return PartialState().on_process(function, process_index)(*args, **kwargs)
        return _inner
    def on_local_process(self, function: Callable[..., Any] = None, local_process_index: int = None):
        if (self is not None) and (local_process_index is not None) and (function is None): return partial(self.on_local_process, local_process_index=local_process_index)
        if function is None:
            if "Accelerator." in self.__qualname__: function = self
            else: raise ValueError("The `on_main_process` decorator must be called with a function on an instantiated `Accelerator` object.")
        def _inner(*args, **kwargs): return PartialState().on_local_process(function, local_process_index)(*args, **kwargs)
        return _inner
    @contextmanager
    def main_process_first(self):
        with self.state.main_process_first(): yield
    @contextmanager
    def local_main_process_first(self):
        with self.state.local_main_process_first(): yield
    @contextmanager
    def no_sync(self, model):
        context = contextlib.nullcontext
        if self.use_distributed: context = getattr(model, "no_sync", context)
        with context(): yield
    @staticmethod
    @contextmanager
    def trigger_sync_in_backward(model):
        if not isinstance(model, torch.nn.parallel.DistributedDataParallel):
            yield
            return
        old_require_backward_grad_sync = model.require_backward_grad_sync
        old_require_forward_param_sync = model.require_forward_param_sync
        model.require_backward_grad_sync = True
        model.require_forward_param_sync = True
        model.reducer.prepare_for_backward([])
        try: yield
        finally:
            model.require_backward_grad_sync = old_require_backward_grad_sync
            model.require_forward_param_sync = old_require_forward_param_sync
    def _do_sync(self):
        if self.gradient_state.sync_with_dataloader and self.gradient_state.end_of_dataloader:
            self.step = 0
            self.gradient_state._set_sync_gradients(True)
        else:
            self.step += 1
            self.gradient_state._set_sync_gradients((self.step % self.gradient_state.num_steps) == 0)
    @property
    def sync_gradients(self): return self.gradient_state.sync_gradients
    @sync_gradients.setter
    def sync_gradients(self, sync_gradients): self.gradient_state.sync_gradients = sync_gradients
    @property
    def gradient_accumulation_steps(self): return self.gradient_state.num_steps
    @gradient_accumulation_steps.setter
    def gradient_accumulation_steps(self, gradient_accumulation_steps): self.gradient_state.plugin_kwargs.update({"num_steps": gradient_accumulation_steps})
    @contextmanager
    def accumulate(self, *models):
        self._do_sync()
        allow_gradient_sync = (self.sync_gradients or (self.use_distributed and self.gradient_state.plugin_kwargs.get("sync_each_batch", False)))
        with contextlib.ExitStack() as cm_stack:
            for m in models: cm_stack.enter_context(contextlib.nullcontext() if allow_gradient_sync else self.no_sync(m))
            yield
    @contextmanager
    def join_uneven_inputs(self, joinables, even_batches=None):
        if self.distributed_type in (DistributedType.MULTI_GPU, DistributedType.MULTI_NPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_XPU):
            dl_even_batches_values = []
            if even_batches is not None:
                iterable_dl_seen = False
                for dl_idx, dl in enumerate(self._dataloaders):
                    if isinstance(dl, DataLoaderDispatcher):
                        iterable_dl_seen = True
                        continue
                    dl_even_batches_values.append((dl_idx, dl.batch_sampler.even_batches))
                    dl.batch_sampler.even_batches = even_batches
            else: even_batches = self.even_batches
            enable_join = False if even_batches else True
            try:
                with Join(joinables, enable=enable_join, throw_on_early_termination=False): yield
            finally:
                for dl_idx, even_batches_value in dl_even_batches_values: self._dataloaders[dl_idx].batch_sampler.even_batches = even_batches_value
        else:
            with contextlib.nullcontext(joinables): yield
    def print(self, *args, **kwargs): self.state.print(*args, **kwargs)
    def _prepare_one(self, obj, first_pass=False, device_placement=None):
        if first_pass:
            if isinstance(obj, torch.utils.data.DataLoader): return self.prepare_data_loader(obj, device_placement=device_placement)
            elif isinstance(obj, torch.nn.Module): return self.prepare_model(obj, device_placement=device_placement)
            elif isinstance(obj, torch.optim.Optimizer):
                optimizer = self.prepare_optimizer(obj, device_placement=device_placement)
                return optimizer
        elif isinstance(obj, LRScheduler):
            scheduler = self.prepare_scheduler(obj)
            return scheduler
        return obj
    def prepare(self, *args, device_placement=None):
        if device_placement is None: device_placement = [None for _ in args]
        elif self.distributed_type in (DistributedType.DEEPSPEED, DistributedType.MEGATRON_LM): raise ValueError("You can't customize device placements with DeepSpeed or Megatron-LM.")
        elif len(device_placement) != len(args): raise ValueError(f"`device_placement` should be a list with {len(args)} elements (the number of objects passed).")
        for obj in args:
            if (isinstance(obj, torch.nn.Module) and self.verify_device_map(obj) and self.distributed_type != DistributedType.NO and os.environ.get("SAPIENS_ACCELERATOR_BYPASS_DEVICE_MAP", "false") != "true"): raise ValueError("You can't train a model that has been loaded with `device_map='auto'` in any distributed mode. Please rerun your script specifying `--num_processes=1` or by launching with `python {{myscript.py}}`.")
        if self.distributed_type == DistributedType.DEEPSPEED:
            model_count = 0
            for obj in args:
                if isinstance(obj, torch.nn.Module): model_count += 1
            if model_count > 1: raise AssertionError("You can't use same `Accelerator()` instance with multiple models when using DeepSpeed")
        if self.distributed_type == DistributedType.XLA:
            model_device, optimizer_device = self._get_devices()
            if model_device is not None and optimizer_device is not None and model_device != optimizer_device: raise ValueError("The model and the optimizer parameters are not on the same device, which probably means you created an optimizer around your model **before** putting on the device. Make sure the line model.to(device) is before the optimizer creation in your script or remove it entirely and use the flag default value for `device_placement` in your `Accelerator` to let it handle that part for you.")
        tpu_should_fix_optimizer = self.device_placement and self.distributed_type == DistributedType.XLA
        if tpu_should_fix_optimizer: old_named_params = self._get_named_parameters(*args)
        if self.distributed_type in [DistributedType.MULTI_CPU, DistributedType.MULTI_XPU, DistributedType.NO]:
            if self.device.type == "cpu" and self.state.use_ipex: args = self._prepare_ipex_or_xpu(*args)
            elif self.device.type == "xpu" and is_xpu_available(): args = self._prepare_ipex_or_xpu(*args)
        if self.fp8_backend == "TE": args = self._prepare_te(*args)
        if self.distributed_type == DistributedType.DEEPSPEED: result = self._prepare_deepspeed(*args)
        elif self.distributed_type == DistributedType.MEGATRON_LM: result = self._prepare_megatron_lm(*args)
        else:
            if self.fp8_backend == "MSAMP": args, device_placement = self._prepare_msamp(*args, device_placement=device_placement)
            result = tuple(self._prepare_one(obj, first_pass=True, device_placement=d) for obj, d in zip(args, device_placement))
            result = tuple(self._prepare_one(obj, device_placement=d) for obj, d in zip(result, device_placement))
        if tpu_should_fix_optimizer:
            new_named_params = self._get_named_parameters(*result)
            mapping = {p: new_named_params[n] for n, p in old_named_params.items()}
            for obj in result:
                if isinstance(obj, torch.optim.Optimizer): obj._switch_parameters(mapping)
        for item in result:
            if any(item in container for container in (self._dataloaders, self._models, self._optimizers, self._schedulers)): item._is_sapiens_accelerator_prepared = True
        return result if len(result) > 1 else result[0]
    def prepare_model(self, model: torch.nn.Module, device_placement: bool = None, evaluation_mode: bool = False):
        if device_placement is None: device_placement = self.device_placement and self.distributed_type != DistributedType.FSDP
        self._models.append(model)
        if (self.verify_device_map(model) and self.distributed_type != DistributedType.NO and os.environ.get("SAPIENS_ACCELERATOR_BYPASS_DEVICE_MAP", "false") != "true"): raise ValueError("You can't train a model that has been loaded with `device_map='auto'` in any distributed mode. Please rerun your script specifying `--num_processes=1` or by launching with `python {{myscript.py}}`.")
        if self.native_amp:
            model._original_forward = model.forward
            autocast_context = get_mixed_precision_context_manager(self.native_amp, self.autocast_handler)
            if self.fp8_backend == "MSAMP" or not hasattr(model.forward, "__func__"):
                model_forward_func = model.forward
                model.forward = convert_outputs_to_fp32(autocast_context(model_forward_func))
            else:
                model_forward_func = model.forward.__func__
                new_forward = autocast_context(model_forward_func)
                model.forward = MethodType(new_forward, model)
                model.forward = MethodType(convert_outputs_to_fp32(model.forward.__func__), model)
        if self.fp8_backend == "TE" and not self.delayed_fp8_autocast: model = apply_fp8_autowrap(model, self.fp8_recipe_handler)
        if (getattr(model, "is_loaded_in_8bit", False) or getattr(model, "is_loaded_in_4bit", False)) and getattr(model, "hf_device_map", False):
            model_devices = set(model.hf_device_map.values())
            if len(model_devices) > 1 and self.distributed_type != DistributedType.NO: raise ValueError("You can't train a model that has been loaded in 8-bit or 4-bit precision on multiple devices in any distributed mode. In order to use 8-bit or 4-bit models that have been loaded across multiple GPUs the solution is to use Naive Pipeline Parallelism. Therefore you should not specify that you are under any distributed regime in your sapiens_accelerator config.")
            elif len(model_devices) == 1:
                current_device = list(model_devices)[0]
                current_device_index = (current_device.index if isinstance(current_device, torch.device) else current_device)
                if torch.device(current_device_index) != self.device:
                    if (self.device.index is not None) or (current_device_index != 0): raise ValueError("You can't train a model that has been loaded in 8-bit or 4-bit precision on a different device than the one you're training on. Make sure you loaded the model on the correct device using for example `device_map={'':torch.cuda.current_device()}` or `device_map={'':torch.xpu.current_device()}`")
            if ("cpu" in model_devices and not is_sapiens_machine_multi_backend_available()) or "disk" in model_devices:
                raise ValueError("You can't train a model that has been loaded in 8-bit or 4-bit precision with CPU or disk offload. If you want train the 8-bit or 4-bit model in CPU, please install sapiens_machine with multi-backend.")
        elif device_placement and not self.verify_device_map(model): model = model.to(self.device)
        if not evaluation_mode:
            if self.distributed_type in (DistributedType.MULTI_GPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_NPU, DistributedType.MULTI_XPU):
                if any(p.requires_grad for p in model.parameters()):
                    kwargs = self.ddp_handler.to_kwargs() if self.ddp_handler is not None else {}
                    if os.environ.get("SAPIENS_ACCELERATOR_BYPASS_DEVICE_MAP", "false") != "true": device_ids, output_device = [self.local_process_index], self.local_process_index
                    else: device_ids, output_device = None, None
                    model = torch.nn.parallel.DistributedDataParallel(model, device_ids=device_ids, output_device=output_device, **kwargs)
                    if self.ddp_handler is not None: self.ddp_handler.register_comm_hook(model)
            elif self.distributed_type == DistributedType.FSDP:
                from torch.distributed.fsdp.fully_sharded_data_parallel import FullyShardedDataParallel as FSDP
                is_type_fsdp = isinstance(model, FSDP) or (is_compiled_module(model) and isinstance(model._orig_mod, FSDP))
                if not is_type_fsdp:
                    self.state.fsdp_plugin.set_auto_wrap_policy(model)
                    fsdp_plugin = self.state.fsdp_plugin
                    kwargs = {"sharding_strategy": fsdp_plugin.sharding_strategy, "cpu_offload": fsdp_plugin.cpu_offload, "auto_wrap_policy": fsdp_plugin.auto_wrap_policy,
                    "mixed_precision": fsdp_plugin.mixed_precision_policy, "sync_module_states": fsdp_plugin.sync_module_states, "backward_prefetch": fsdp_plugin.backward_prefetch,
                    "forward_prefetch": fsdp_plugin.forward_prefetch, "use_orig_params": fsdp_plugin.use_orig_params, "param_init_fn": fsdp_plugin.param_init_fn,
                    "ignored_modules": fsdp_plugin.ignored_modules, "limit_all_gathers": fsdp_plugin.limit_all_gathers, "device_id": self.device}
                    model = FSDP(model, **kwargs)
                    if fsdp_plugin.activation_checkpointing:
                        from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import (CheckpointImpl, apply_activation_checkpointing, checkpoint_wrapper)
                        apply_activation_checkpointing(model, checkpoint_wrapper_fn=functools.partial(checkpoint_wrapper, checkpoint_impl=CheckpointImpl.NO_REENTRANT), auto_wrap_policy=fsdp_plugin.auto_wrap_policy)
                if self.mixed_precision != "no":
                    upcasted_log = []
                    for module in FSDP.fsdp_modules(model):
                        if not module._has_params: continue
                        param = module._flat_param
                        if (param.dtype != torch.float32 and param.device != torch.device("meta") and param.requires_grad):
                            name_param_log = (module.module.__class__.__name__, ", ".join(module._flat_param._fqns))
                            if name_param_log not in upcasted_log: upcasted_log.append(name_param_log)
                            param.data = param.data.to(torch.float32)
                            module._handle._orig_param_dtype = torch.float32
                if len(self._models) > 1 and (self._models[-2] is self._models[-1]): del self._models[-2]
                self._models[-1] = model
            elif self.distributed_type == DistributedType.MULTI_CPU:
                kwargs = self.ddp_handler.to_kwargs() if self.ddp_handler is not None else {}
                model = torch.nn.parallel.DistributedDataParallel(model, **kwargs)
                if self.ddp_handler is not None: self.ddp_handler.register_comm_hook(model)
            elif self.distributed_type == DistributedType.XLA and self.state.fork_launched: model = xmp.MpModelWrapper(model).to(self.device)
        if self.delayed_fp8_autocast: model = apply_fp8_autowrap(model, self.fp8_recipe_handler)
        if self.state.dynamo_plugin.backend != DynamoBackend.NO and not is_compiled_module(model):
            if not is_torch_version(">=", "2.0"): raise ValueError("Using `torch.compile` requires PyTorch 2.0 or higher.")
            model = torch.compile(model, **self.state.dynamo_plugin.to_kwargs())
        return model
    def _prepare_te(self, *args):
        if not is_transformer_engine_available(): raise ImportError("`transformer_engine` was not found on your system. Please ensure that `transformer_engine` is installed")
        model, optimizer = None, None
        num_models, num_optimizers = 0, 0
        result = [obj for obj in args]
        for obj in result:
            if isinstance(obj, torch.nn.Module):
                model = obj
                num_models += 1
            elif isinstance(obj, (torch.optim.Optimizer)):
                optimizer = obj
                num_optimizers += 1
        if optimizer is None and model is None: return result
        elif optimizer is None or model is None: raise ValueError("You must pass a model and an optimizer together to `sapiens_accelerator.prepare()` when using TransformerEngine.")
        elif num_models > 1 or num_optimizers > 1: raise ValueError(f"You can't use multiple models ({num_models}) or optimizers {num_optimizers} with TransformerEngine.")
        old_named_params = self._get_named_parameters(model)
        with torch.no_grad(): convert_model(model)
        new_named_params = self._get_named_parameters(model)
        mapping = {p: new_named_params[n] for n, p in old_named_params.items()}
        for param_group in optimizer.param_groups: param_group["params"] = [mapping[p] for p in param_group["params"]]
        return result
    def _prepare_deepspeed(self, *args):
        import deepspeed
        ds_initialize = deepspeed.initialize
        if self.fp8_backend == "MSAMP":
            from msamp import deepspeed as msamp_deepspeed
            ds_initialize = msamp_deepspeed.initialize
        deepspeed_plugin = self.deepspeed_plugin
        is_dataloader_present = any(isinstance(obj, torch.utils.data.DataLoader) for obj in args)
        result = [self._prepare_one(obj, first_pass=True) if isinstance(obj, torch.utils.data.DataLoader) else obj for obj in args]
        if deepspeed_plugin.is_auto("train_micro_batch_size_per_gpu"):
            if is_dataloader_present:
                batch_sizes = [obj.batch_size for obj in args if hasattr(obj, "batch_size")]
                if any(bs is None for bs in batch_sizes): raise ValueError("At least one of the dataloaders passed to `sapiens_accelerator.prepare()` has `None` as batch size. Please set an integer value in `train_micro_batch_size_per_gpu` in the deepspeed config file or assign integer value to `AcceleratorState().deepspeed_plugin.deepspeed_config['train_micro_batch_size_per_gpu']`.")
                if self.split_batches: batch_sizes = [batch_size // self.num_processes for batch_size in batch_sizes]
                batch_size_per_device = min(batch_sizes) if deepspeed_plugin.is_train_batch_min else max(batch_sizes)
            else: raise ValueError("When using DeepSpeed, `sapiens_accelerator.prepare()` requires you to pass at least one of training or evaluation dataloaders with `batch_size` attribute returning an integer value or alternatively set an integer value in `train_micro_batch_size_per_gpu` in the deepspeed config file or assign integer value to `AcceleratorState().deepspeed_plugin.deepspeed_config['train_micro_batch_size_per_gpu']`.")
        else: batch_size_per_device = deepspeed_plugin.get_value("train_micro_batch_size_per_gpu")
        deepspeed_plugin.fill_match("gradient_accumulation_steps", must_match=False, gradient_accumulation_steps=self.gradient_accumulation_steps)
        config_kwargs = {"gradient_clipping": 1.0, "zero_optimization.stage3_gather_16bit_weights_on_model_save": False}
        if batch_size_per_device is not None:
            config_kwargs["train_micro_batch_size_per_gpu"] = batch_size_per_device
            config_kwargs["train_batch_size"] = (batch_size_per_device * deepspeed_plugin.get_value("gradient_accumulation_steps") * self.num_processes)
        model = None
        optimizer = None
        scheduler = None
        for obj in result:
            if isinstance(obj, torch.nn.Module): model = obj
            elif isinstance(obj, (torch.optim.Optimizer, DummyOptim)): optimizer = obj
            elif (isinstance(obj, (LRScheduler, DummyScheduler))) or (type(obj).__name__ in deepspeed.runtime.lr_schedules.VALID_LR_SCHEDULES): scheduler = obj
        if optimizer is not None:
            if "optimizer" in deepspeed_plugin.deepspeed_config and not isinstance(optimizer, (DummyOptim)): raise ValueError("You cannot specify an optimizer in the config file and in the code at the same time. Please remove the optimizer from the config file or create `sapiens_accelerator.utils.DummyOptim` in the code.")
            elif "optimizer" not in deepspeed_plugin.deepspeed_config and isinstance(optimizer, (DummyOptim)): raise ValueError("You cannot create a `DummyOptim` without specifying an optimizer in the config file.")
            if isinstance(optimizer, (torch.optim.Optimizer)): deepspeed_plugin.deepspeed_config["zero_allow_untested_optimizer"] = True
        if scheduler is not None:
            if "scheduler" in deepspeed_plugin.deepspeed_config and not isinstance(scheduler, (DummyScheduler)): raise ValueError("You cannot specify a scheduler in the config file and in the code at the same time. Please remove the scheduler from the config file or create `sapiens_accelerator.utils.DummyScheduler` in the code.")
            elif ("scheduler" not in deepspeed_plugin.deepspeed_config and isinstance(scheduler, (DummyScheduler)) and scheduler.lr_scheduler_callable is None): raise ValueError("Either specify a scheduler in the config file or pass in the `lr_scheduler_callable` parameter when using `sapiens_accelerator.utils.DummyScheduler`.")
        if optimizer is not None and scheduler is not None:
            if isinstance(optimizer, (DummyOptim)) and not isinstance(scheduler, (DummyScheduler)): raise ValueError("You can only specify `sapiens_accelerator.utils.DummyScheduler` in the code when using `sapiens_accelerator.utils.DummyOptim`.")
        if model is not None:
            if getattr(self.fp8_recipe_handler, "backend", None) == "TE": model = apply_fp8_autowrap(model, self.fp8_recipe_handler)
            deepspeed_plugin.set_moe_leaf_modules(model)
            hidden_size_based_keys = ["zero_optimization.reduce_bucket_size", "zero_optimization.stage3_prefetch_bucket_size", "zero_optimization.stage3_param_persistence_threshold"]
            hidden_size_auto_keys = [x for x in hidden_size_based_keys if deepspeed_plugin.is_auto(x)]
            if len(hidden_size_auto_keys) > 0:
                reasoning = ("therefore it's not possible to automatically fill out the following `auto` entries " + f"in the DeepSpeed config file: {hidden_size_auto_keys}. You can fix that by replacing " + "`auto` values for these keys with an integer value of your choice.")
                if not hasattr(model, "config"): raise ValueError("Can't find `model.config` entry, " + reasoning)
                if hasattr(model.config, "hidden_size"): hidden_size = model.config.hidden_size
                elif hasattr(model.config, "hidden_sizes"): hidden_size = max(model.config.hidden_sizes)
                else: raise ValueError("Can find neither `model.config.hidden_size` nor `model.config.hidden_sizes`, " + reasoning)
                config_kwargs.update({"zero_optimization.reduce_bucket_size": hidden_size * hidden_size, "zero_optimization.stage3_prefetch_bucket_size": int(0.9 * hidden_size * hidden_size),
                "zero_optimization.stage3_param_persistence_threshold": 10 * hidden_size})
            if isinstance(optimizer, (DummyOptim)): config_kwargs.update({"optimizer.params.lr": optimizer.lr, "optimizer.params.weight_decay": optimizer.weight_decay})
            if isinstance(scheduler, (DummyScheduler)) and scheduler.lr_scheduler_callable is None:
                max_lr = (getattr(scheduler.optimizer, "lr", None) if getattr(scheduler.optimizer, "defaults", None) is None else scheduler.optimizer.defaults["lr"])
                config_kwargs.update({"scheduler.params.warmup_min_lr": 0, "scheduler.params.warmup_max_lr": max_lr, "scheduler.params.warmup_num_steps": scheduler.warmup_num_steps})
                if scheduler.total_num_steps is not None: config_kwargs["scheduler.params.total_num_steps"] = (math.ceil(scheduler.total_num_steps / self.num_processes) if not self.split_batches else scheduler.total_num_steps)
            deepspeed_plugin.deepspeed_config_process(must_match=False, **config_kwargs)
            self.deepspeed_config = deepspeed_plugin.deepspeed_config
            kwargs = dict(model=model, config_params=self.deepspeed_config)
            if optimizer is not None:
                if isinstance(optimizer, (DummyOptim)):
                    kwargs["model_parameters"] = optimizer.params
                    if isinstance(scheduler, (DummyScheduler)) and scheduler.lr_scheduler_callable is not None: kwargs["lr_scheduler"] = scheduler.lr_scheduler_callable
                else:
                    if self.deepspeed_config["zero_optimization"].get("offload_optimizer", {}).get("device", "none") != "none" and self.deepspeed_config.get("zero_force_ds_cpu_optimizer", True):
                        from deepspeed.ops.adam import DeepSpeedCPUAdam
                        defaults = {k: v for k, v in optimizer.defaults.items() if k in ["lr", "weight_decay"]}
                        optimizer = DeepSpeedCPUAdam(optimizer.param_groups, **defaults)
                    kwargs["optimizer"] = optimizer
                    if scheduler is not None:
                        if type(scheduler).__name__ in deepspeed.runtime.lr_schedules.VALID_LR_SCHEDULES: kwargs["lr_scheduler"] = scheduler
            engine, optimizer, _, lr_scheduler = ds_initialize(**kwargs)
            if optimizer is not None: optimizer = DeepSpeedOptimizerWrapper(optimizer)
            if scheduler is not None:
                if lr_scheduler is None: scheduler = SapiensAcceleratordScheduler(scheduler, optimizer, step_with_optimizer=self.step_scheduler_with_optimizer, split_batches=self.split_batches)
                else: scheduler = DeepSpeedSchedulerWrapper(lr_scheduler, optimizer)
            for i in range(len(result)):
                if isinstance(result[i], torch.nn.Module): result[i] = engine
                elif isinstance(result[i], (torch.optim.Optimizer, DummyOptim)): result[i] = optimizer
                elif (isinstance(result[i], (LRScheduler, DummyScheduler))) or (type(result[i]).__name__ in deepspeed.runtime.lr_schedules.VALID_LR_SCHEDULES): result[i] = scheduler
            if self.deepspeed_engine_wrapped is None: self.deepspeed_engine_wrapped = DeepSpeedEngineWrapper(engine)
            self._models.append(engine)
            if optimizer is not None: self._optimizers.append(optimizer)
            if scheduler is not None: self._schedulers.append(scheduler)
        return tuple(result)
    def _prepare_megatron_lm(self, *args):
        megatron_lm_plugin = self.state.megatron_lm_plugin
        micro_batch_size = None
        if not megatron_lm_plugin.megatron_dataset_flag:
            batch_sizes = [obj.batch_size for obj in args if hasattr(obj, "batch_size")]
            if len(batch_sizes) == 0: raise ValueError("You must specify a training or evaluation dataloader in `sapiens_accelerator.prepare()` when using Megatron-LM.")
            micro_batch_size = min(batch_sizes) if megatron_lm_plugin.is_train_batch_min else max(batch_sizes)
        else:
            for obj in args:
                if isinstance(obj, MegatronLMDummyDataLoader):
                    micro_batch_size = obj.dataset_args["micro_batch_size"]
                    break
        if micro_batch_size is not None:
            dp_degree = self.num_processes // (megatron_lm_plugin.tp_degree * megatron_lm_plugin.pp_degree)
            megatron_lm_plugin.set_training_args(micro_batch_size, dp_degree)
        else: raise ValueError("When you do not pass the dataloader parameter, the `data_parallel_size`, `micro_batch_size`, and `global_batch_size` megatron parameters will not be updated.")
        model = None
        optimizer = None
        scheduler = None
        batch_data = None
        for obj in args:
            if isinstance(obj, torch.utils.data.DataLoader) and batch_data is None: batch_data = next(iter(obj))
            elif isinstance(obj, torch.nn.Module): model = obj
            elif isinstance(obj, (torch.optim.Optimizer)): optimizer = obj
            elif isinstance(obj, (LRScheduler, MegatronLMDummyScheduler)): scheduler = obj
        if model is not None: megatron_lm_plugin.set_network_size_args(model, batch_data)
        if optimizer is not None: megatron_lm_plugin.set_optimizer_type(optimizer)
        if scheduler is not None:
            if not isinstance(scheduler, MegatronLMDummyScheduler): raise ValueError("You can't use a custom scheduler with Megatron-LM. Please use the `sapiens_accelerator.utils.MegatronLMDummyScheduler` instead.")
            megatron_lm_plugin.set_scheduler_args(scheduler)
        megatron_lm_initialize(self, args_defaults=megatron_lm_plugin.megatron_lm_default_args)
        (model, optimizer, scheduler) = megatron_lm_prepare_model_optimizer_scheduler(self)
        self.wait_for_everyone()
        counter = 0
        result = []
        for obj in args:
            if isinstance(obj, torch.utils.data.DataLoader):
                result.append(megatron_lm_prepare_data_loader(self, obj))
                counter += 1
            elif isinstance(obj, MegatronLMDummyDataLoader):
                if counter == 0:
                    obj.set_megatron_data_args()
                    dataloaders = megatron_lm_prepare_data_loader(self, obj)
                result.append(dataloaders[counter])
                counter += 1
            else: result.append(obj)
        if model is not None: model = MegatronEngine(self, model, optimizer, scheduler)
        if optimizer is not None: optimizer = MegatronLMOptimizerWrapper(optimizer)
        if scheduler is not None: scheduler = MegatronLMSchedulerWrapper(scheduler, optimizer)
        for i in range(len(result)):
            if isinstance(result[i], torch.nn.Module): result[i] = model
            elif isinstance(result[i], torch.optim.Optimizer): result[i] = optimizer
            elif isinstance(result[i], MegatronLMDummyScheduler): result[i] = scheduler
        if model is not None:
            self._models.append(model)
            if len(self._models) > 1: raise AssertionError("You can't use same `Accelerator()` instance with multiple models when using Megatron-LM")
        if optimizer is not None: self._optimizers.append(optimizer)
        if scheduler is not None: self._schedulers.append(scheduler)
        return tuple(result)
    def _prepare_ipex_or_xpu(self, *args):
        if self.state.use_ipex:
            if not is_ipex_available(): raise ImportError("IPEX is not installed or IPEX's version does not match current PyTorch version. Please refer to https://github.com/intel/intel-extension-for-pytorch.")
        model = None
        optimizer = None
        result = [obj for obj in args]
        for obj in result:
            if isinstance(obj, torch.nn.Module):
                model = obj
                model.train()
            elif isinstance(obj, (torch.optim.Optimizer)): optimizer = obj
        if optimizer is not None and model is not None:
            dtype = torch.bfloat16 if self.state.mixed_precision == "bf16" else None
            if self.device.type == "xpu" and model.device.type == "cpu": model = model.to(self.device)
            if is_ipex_available():
                import intel_extension_for_pytorch as ipex
                model, optimizer = ipex.optimize(model, optimizer=optimizer, dtype=dtype, inplace=True, level="O1")
        for i in range(len(result)):
            if isinstance(result[i], torch.nn.Module): result[i] = model
            elif isinstance(result[i], (torch.optim.Optimizer)): result[i] = optimizer
        return tuple(result)
    def _prepare_msamp(self, *args, device_placement):
        if not is_msamp_available(): raise ImportError("MS-AMP was not found on your system. Please ensure that MS-AMP is available or choose `'te'` as the backend for FP8 mixed precision training.")
        import msamp
        model, optimizer = None, None
        optimizer_index = None
        num_models, num_optimizers = 0, 0
        result = [obj for obj in args]
        for i, obj in enumerate(result):
            if isinstance(obj, torch.nn.Module):
                model = obj
                num_models += 1
            elif isinstance(obj, (torch.optim.Optimizer)):
                optimizer = obj
                optimizer_index = i
                num_optimizers += 1
        if optimizer is None and model is None: return result, device_placement
        elif optimizer is None or model is None: raise ValueError("You must pass a model and an optimizer together to `sapiens_accelerator.prepare()` when using MS-AMP.")
        elif num_models > 1 or num_optimizers > 1: raise ValueError(f"You can't use multiple models ({num_models}) or optimizers {num_optimizers} with MS-AMP.")
        else: model, optimizer = msamp.initialize(model, optimizer, opt_level=self.fp8_recipe_handler.opt_level)
        for i in range(len(result)):
            if isinstance(result[i], torch.nn.Module): result[i] = model
            elif isinstance(result[i], (torch.optim.Optimizer)): result[i] = optimizer
        if optimizer_index is not None: device_placement[optimizer_index] = False
        return tuple(result), device_placement
    def prepare_data_loader(self, data_loader: torch.utils.data.DataLoader, device_placement=None, slice_fn_for_dispatch=None):
        if getattr(data_loader, "_is_sapiens_accelerator_prepared", False):
            if data_loader not in self._dataloaders: self._dataloaders.append(data_loader)
            return data_loader
        if device_placement is None: device_placement = self.device_placement if self.distributed_type != DistributedType.XLA else False
        prepared_data_loader = prepare_data_loader(data_loader, self.device, num_processes=self.num_processes, process_index=self.process_index,
        split_batches=self.split_batches, put_on_device=device_placement, rng_types=self.rng_types.copy(), dispatch_batches=self.dispatch_batches,
        even_batches=self.even_batches, slice_fn_for_dispatch=slice_fn_for_dispatch, use_seedable_sampler=self.use_seedable_sampler,
        non_blocking=self.non_blocking, use_stateful_dataloader=self.use_stateful_dataloader)
        self._dataloaders.append(prepared_data_loader)
        return prepared_data_loader
    def prepare_optimizer(self, optimizer: torch.optim.Optimizer, device_placement=None):
        if is_lomo_available():
            from lomo_optim import AdaLomo, Lomo
            self.has_lomo_optimizer |= isinstance(optimizer, (Lomo, AdaLomo))
        if getattr(optimizer, "_is_sapiens_accelerator_prepared", False):
            if optimizer not in self._optimizers: self._optimizers.append(optimizer)
            return optimizer
        if device_placement is None: device_placement = self.device_placement
        scaler = None if self.fp8_backend == "MSAMP" else self.scaler
        optimizer = SapiensAcceleratordOptimizer(optimizer, device_placement=device_placement, scaler=scaler)
        self._optimizers.append(optimizer)
        return optimizer
    def prepare_scheduler(self, scheduler: LRScheduler):
        if getattr(scheduler, "_is_sapiens_accelerator_prepared", False):
            if scheduler not in self._schedulers: self._schedulers.append(scheduler)
            return scheduler
        optimizer = self._optimizers
        for opt in self._optimizers:
            if getattr(scheduler, "optimizer", None) == opt.optimizer:
                optimizer = opt
                break
        scheduler = SapiensAcceleratordScheduler(scheduler, optimizer, step_with_optimizer=self.step_scheduler_with_optimizer, split_batches=self.split_batches)
        self._schedulers.append(scheduler)
        return scheduler
    def backward(self, loss, **kwargs):
        learning_rate = kwargs.get("learning_rate")
        if self.distributed_type != DistributedType.DEEPSPEED: loss = loss / self.gradient_accumulation_steps
        if self.distributed_type == DistributedType.DEEPSPEED: self.deepspeed_engine_wrapped.backward(loss, **kwargs)
        elif self.distributed_type == DistributedType.MEGATRON_LM: return
        elif self.scaler is not None: self.scaler.scale(loss).backward(**kwargs)
        elif learning_rate is not None and self.has_lomo_optimizer: self.lomo_backward(loss, learning_rate)
        else: loss.backward(**kwargs)
    def set_trigger(self): self.flag_tensor = torch.tensor(1, device=self.device)
    def check_trigger(self):
        if self.flag_tensor is None: self.flag_tensor = torch.tensor(0, device=self.device)
        flag_tensor = self.reduce(self.flag_tensor)
        if flag_tensor.item() >= 1:
            self.flag_tensor = torch.tensor(0, device=self.device)
            return True
        return False
    def unscale_gradients(self, optimizer=None):
        if self.native_amp and self.mixed_precision == "fp16":
            if optimizer is None: optimizer = self._optimizers
            elif not isinstance(optimizer, (tuple, list)): optimizer = [optimizer]
            for opt in optimizer:
                while isinstance(opt, SapiensAcceleratordOptimizer): opt = opt.optimizer
                self.scaler.unscale_(opt)
    def clip_grad_norm_(self, parameters, max_norm, norm_type=2):
        if self.distributed_type == DistributedType.FSDP:
            self.unscale_gradients()
            parameters = [p for p in parameters]
            for model in self._models:
                if parameters == [p for p in model.parameters()]: return model.clip_grad_norm_(max_norm, norm_type)
        elif self.distributed_type == DistributedType.DEEPSPEED: return None
        elif self.distributed_type == DistributedType.XLA:
            for acc_opt in self._optimizers:
                if not acc_opt.gradient_state.is_xla_gradients_synced:
                    opt = acc_opt
                    while isinstance(opt, SapiensAcceleratordOptimizer): opt = opt.optimizer
                    gradients = xm._fetch_gradients(opt)
                    xm.all_reduce("sum", gradients, scale=1.0 / self.num_processes)
                    acc_opt.gradient_state.is_xla_gradients_synced = True
            if os.environ.get("SAPIENS_ACCELERATOR_USE_FSDP", "false") == "true":
                self.unscale_gradients()
                parameters = [p for p in parameters]
                for model in self._models:
                    if parameters == [p for p in model.parameters()]: return model.clip_grad_norm_(max_norm, norm_type)
        self.unscale_gradients()
        return torch.nn.utils.clip_grad_norm_(parameters, max_norm, norm_type=norm_type)
    def clip_grad_value_(self, parameters, clip_value):
        if self.distributed_type in [DistributedType.DEEPSPEED, DistributedType.FSDP]: raise Exception("DeepSpeed and FSDP  do not support `clip_grad_value_`. Use `clip_grad_norm_` instead.")
        self.unscale_gradients()
        torch.nn.utils.clip_grad_value_(parameters, clip_value)
    def gather(self, tensor): return gather(tensor)
    def gather_for_metrics(self, input_data, use_gather_object=False):
        try:
            recursively_apply(lambda x: x, input_data, error_on_other_type=True)
            all_tensors = True
        except TypeError: all_tensors = False
        use_gather_object = use_gather_object or not all_tensors
        if use_gather_object: data = gather_object(input_data)
        else: data = self.gather(input_data)
        try:
            if self.gradient_state.end_of_dataloader:
                if self.gradient_state.remainder == -1: return data
                elif self.gradient_state.remainder > 0:
                    def _adjust_samples(tensor): return tensor[: self.gradient_state.remainder]
                    if use_gather_object: return _adjust_samples(data)
                    else: return recursively_apply(_adjust_samples, data)
                else: return data
            else: return data
        except Exception: return data
    def reduce(self, tensor, reduction="sum", scale=1.0): return reduce(tensor, reduction, scale)
    def pad_across_processes(self, tensor, dim=0, pad_index=0, pad_first=False): return pad_across_processes(tensor, dim=dim, pad_index=pad_index, pad_first=pad_first)
    def unwrap_model(self, model, keep_fp32_wrapper: bool = True): return extract_model_from_parallel(model, keep_fp32_wrapper)
    def wait_for_everyone(self): wait_for_everyone()
    @on_main_process
    def init_trackers(self, project_name: str, config: dict | None = None, init_kwargs: dict | None = {}):
        for tracker in self.log_with:
            if issubclass(type(tracker), GeneralTracker): self.trackers.append(tracker)
            else:
                tracker_init = LOGGER_TYPE_TO_CLASS[str(tracker)]
                if tracker_init.requires_logging_directory: self.trackers.append(tracker_init(project_name, self.logging_dir, **init_kwargs.get(str(tracker), {})))
                else: self.trackers.append(tracker_init(project_name, **init_kwargs.get(str(tracker), {})))
        if config is not None:
            for tracker in self.trackers: tracker.store_init_configuration(config)
    def get_tracker(self, name: str, unwrap: bool = False):
        if len(self.trackers) > 0:
            for tracker in self.trackers:
                if tracker.name == name: return tracker.tracker if unwrap else tracker
            raise ValueError(f"{name} is not an available tracker stored inside the `Accelerator`.")
        return GeneralTracker(_blank=True)
    @on_main_process
    def log(self, values: dict, step: int | None = None, log_kwargs: dict | None = {}):
        for tracker in self.trackers: tracker.log(values, step=step, **log_kwargs.get(tracker.name, {}))
    def end_training(self):
        for tracker in self.trackers: tracker.finish()
        self.state.destroy_process_group()
    def save(self, obj, f, safe_serialization=False): save(obj, f, save_on_each_node=self.project_configuration.save_on_each_node, safe_serialization=safe_serialization)
    def save_model(self, model: torch.nn.Module, save_directory: Union[str, os.PathLike], max_shard_size: Union[int, str] = "10GB", safe_serialization: bool = True):
        if os.path.isfile(save_directory): return
        os.makedirs(save_directory, exist_ok=True)
        if any([module._hf_hook.offload for module in model.modules() if hasattr(module, "_hf_hook") and isinstance(module._hf_hook, AlignDevicesHook)]): state_dict = get_state_dict_offloaded_model(model)
        else:
            if any(param.device == torch.device("meta") for param in model.parameters()): raise RuntimeError("You can't save the model since some parameters are on the meta device.")
            state_dict = self.get_state_dict(model)
        if safe_serialization: state_dict = clean_state_dict_for_safetensors(state_dict)
        weights_name = SAFE_WEIGHTS_NAME if safe_serialization else WEIGHTS_NAME
        filename_pattern = SAFE_WEIGHTS_PATTERN_NAME if safe_serialization else WEIGHTS_PATTERN_NAME
        state_dict_split = split_torch_state_dict_into_shards(state_dict, filename_pattern=filename_pattern, max_shard_size=max_shard_size)
        for filename in os.listdir(save_directory):
            full_filename = os.path.join(save_directory, filename)
            weights_no_suffix = weights_name.replace(".bin", "")
            filename_no_suffix = filename.replace(".bin", "")
            reg = re.compile(r"(.*?)-\d{5}-of-\d{5}")
            if (filename.startswith(weights_no_suffix) and os.path.isfile(full_filename) and filename not in state_dict_split.filename_to_tensors.keys()
            and reg.fullmatch(filename_no_suffix) is not None and PartialState().is_main_process): os.remove(full_filename)
        for filename, tensors in state_dict_split.filename_to_tensors.items():
            shard = {tensor: state_dict[tensor] for tensor in tensors}
            self.save(shard, os.path.join(save_directory, filename), safe_serialization=safe_serialization)
        if state_dict_split.is_sharded:
            index = {"metadata": state_dict_split.metadata, "weight_map": state_dict_split.tensor_to_filename}
            save_index_file = SAFE_WEIGHTS_INDEX_NAME if safe_serialization else WEIGHTS_INDEX_NAME
            save_index_file = os.path.join(save_directory, save_index_file)
            with open(save_index_file, "w", encoding="utf-8") as f:
                content = json.dumps(index, indent=2, sort_keys=True) + "\n"
                f.write(content)
        else: path_to_weights = os.path.join(save_directory, WEIGHTS_NAME)
    def register_save_state_pre_hook(self, hook: Callable[..., None]) -> hooks.RemovableHandle:
        handle = hooks.RemovableHandle(self._save_model_state_pre_hook)
        self._save_model_state_pre_hook[handle.id] = hook
        return handle
    def save_state(self, output_dir: str = None, safe_serialization: bool = True, **save_model_func_kwargs):
        if self.project_configuration.automatic_checkpoint_naming: output_dir = os.path.join(self.project_dir, "checkpoints")
        os.makedirs(output_dir, exist_ok=True)
        if self.project_configuration.automatic_checkpoint_naming:
            folders = [os.path.join(output_dir, folder) for folder in os.listdir(output_dir)]
            if (self.project_configuration.total_limit is not None and (len(folders) + 1 > self.project_configuration.total_limit) and self.is_main_process):
                def _inner(folder): return list(map(int, re.findall(r"[\/]?([0-9]+)(?=[^\/]*$)", folder)))[0]
                folders.sort(key=_inner)
                for folder in folders[: len(folders) + 1 - self.project_configuration.total_limit]: shutil.rmtree(folder)
            output_dir = os.path.join(output_dir, f"checkpoint_{self.save_iteration}")
            if os.path.exists(output_dir): raise ValueError(f"Checkpoint directory {output_dir} ({self.save_iteration}) already exists. Please manually override `self.save_iteration` with what iteration to start with.")
            self.wait_for_everyone()
        os.makedirs(output_dir, exist_ok=True)
        if self.distributed_type == DistributedType.XLA: xm.mark_step()
        weights = []
        for i, model in enumerate(self._models):
            if self.distributed_type == DistributedType.FSDP: save_fsdp_model(self.state.fsdp_plugin, self, model, output_dir, i)
            elif self.distributed_type == DistributedType.DEEPSPEED:
                ckpt_id = f"{MODEL_NAME}" if i == 0 else f"{MODEL_NAME}_{i}"
                model.save_checkpoint(output_dir, ckpt_id, **save_model_func_kwargs)
            elif self.distributed_type == DistributedType.MEGATRON_LM: model.save_checkpoint(output_dir)
            else: weights.append(self.get_state_dict(model, unwrap=False))
        optimizers = []
        if self.distributed_type == DistributedType.FSDP:
            for i, opt in enumerate(self._optimizers): save_fsdp_optimizer(self.state.fsdp_plugin, self, opt, self._models[i], output_dir, i)
        elif self.distributed_type not in [DistributedType.DEEPSPEED, DistributedType.MEGATRON_LM]: optimizers = self._optimizers
        schedulers = []
        if self.distributed_type == DistributedType.DEEPSPEED:
            for i, scheduler in enumerate(self._schedulers):
                if isinstance(scheduler, DeepSpeedSchedulerWrapper): continue
                schedulers.append(scheduler)
        elif self.distributed_type not in [DistributedType.MEGATRON_LM]: schedulers = self._schedulers
        dataloaders = self._dataloaders
        for hook in self._save_model_state_pre_hook.values(): hook(self._models, weights, output_dir)
        save_location = save_accelerator_state(output_dir, weights, optimizers, schedulers, dataloaders, self.state.process_index, self.step,
        self.scaler, save_on_each_node=self.project_configuration.save_on_each_node, safe_serialization=safe_serialization)
        for i, obj in enumerate(self._custom_objects): save_custom_state(obj, output_dir, i, save_on_each_node=self.project_configuration.save_on_each_node)
        self.project_configuration.iteration += 1
        return save_location
    def register_load_state_pre_hook(self, hook: Callable[..., None]) -> hooks.RemovableHandle:
        handle = hooks.RemovableHandle(self._load_model_state_pre_hook)
        self._load_model_state_pre_hook[handle.id] = hook
        return handle
    def load_state(self, input_dir: str = None, **load_model_func_kwargs):
        if input_dir is not None:
            input_dir = os.path.expanduser(input_dir)
            if not os.path.isdir(input_dir): raise ValueError(f"Tried to find {input_dir} but folder does not exist")
        elif self.project_configuration.automatic_checkpoint_naming:
            input_dir = os.path.join(self.project_dir, "checkpoints")
            folders = [os.path.join(input_dir, folder) for folder in os.listdir(input_dir)]
            def _inner(folder): return list(map(int, re.findall(r"[\/]?([0-9]+)(?=[^\/]*$)", folder)))[0]
            folders.sort(key=_inner)
            input_dir = folders[-1]
        else: raise ValueError("No input_dir provided and automatic checkpoint naming is disabled.")
        models = []
        for i, model in enumerate(self._models):
            if self.distributed_type == DistributedType.FSDP: load_fsdp_model(self.state.fsdp_plugin, self, model, input_dir, i)
            elif self.distributed_type == DistributedType.DEEPSPEED:
                ckpt_id = f"{MODEL_NAME}" if i == 0 else f"{MODEL_NAME}_{i}"
                model.load_checkpoint(input_dir, ckpt_id, **load_model_func_kwargs)
            elif self.distributed_type == DistributedType.MEGATRON_LM: model.load_checkpoint(input_dir)
            else: models.append(model)
        optimizers = []
        if self.distributed_type == DistributedType.FSDP:
            for i, opt in enumerate(self._optimizers): load_fsdp_optimizer(self.state.fsdp_plugin, self, opt, self._models[i], input_dir, i)
        elif self.distributed_type not in [DistributedType.DEEPSPEED, DistributedType.MEGATRON_LM]: optimizers = self._optimizers
        schedulers = []
        if self.distributed_type == DistributedType.DEEPSPEED:
            for i, scheduler in enumerate(self._schedulers):
                if isinstance(scheduler, DeepSpeedSchedulerWrapper): continue
                schedulers.append(scheduler)
        elif self.distributed_type not in [DistributedType.MEGATRON_LM]: schedulers = self._schedulers
        dataloaders = self._dataloaders
        for hook in self._load_model_state_pre_hook.values(): hook(models, input_dir)
        map_location = load_model_func_kwargs.pop("map_location", None)
        if map_location is None:
            if self.num_processes > 1 and self.distributed_type in (DistributedType.MULTI_GPU, DistributedType.MULTI_MLU, DistributedType.MULTI_MUSA, DistributedType.MULTI_NPU): map_location = "on_device"
            else: map_location = "cpu"
        override_attributes = load_accelerator_state(input_dir, models, optimizers, schedulers, dataloaders, self.state.process_index, self.scaler, map_location, **load_model_func_kwargs)
        if "step" in override_attributes: self.step = override_attributes["step"]
        custom_checkpoints = [f for f in os.listdir(input_dir) if re.search(r"^custom_checkpoint_\d+\.pkl$", f) is not None]
        if len(custom_checkpoints) != len(self._custom_objects):
            err = (f"Number of custom checkpoints in folder {input_dir} does not match the number of registered objects:")
            err += f"\n\tFound checkpoints: {len(custom_checkpoints)}"
            err += f"\n\tRegistered objects: {len(self._custom_objects)}\n"
            err += "Please make sure to only load checkpoints from folders that were created with the same set of registered objects,"
            err += "or avoid using `custom_checkpoint` in the filename for files in that same directory and load them in manually."
            raise RuntimeError(err)
        else:
            for index, obj in enumerate(self._custom_objects): load_custom_state(obj, input_dir, index)
    def free_memory(self, *objects):
        if hasattr(self, "deepspeed_engine_wrapped"):
            if self.deepspeed_engine_wrapped is not None: self.deepspeed_engine_wrapped.engine.destroy()
            self.deepspeed_engine_wrapped = None
        objects = release_memory(*objects)
        self._schedulers = []
        self._optimizers = []
        self._models = []
        self._dataloaders = []
        self.step = 0
        return objects
    def clear(self, *objects): return self.free_memory(*objects)
    def _get_named_parameters(self, *args):
        named_parameters = {}
        for obj in args:
            if isinstance(obj, torch.nn.Module):
                obj = extract_model_from_parallel(obj)
                named_parameters.update({n: p for n, p in obj.named_parameters()})
        return named_parameters
    def _get_devices(self, *args):
        model_device = None
        optimizer_device = None
        for obj in args:
            if isinstance(obj, torch.nn.Module):
                for param in obj.parameters():
                    model_device = param.device
                    break
            if isinstance(obj, torch.optim.Optimizer):
                for param_group in obj.param_groups:
                    if len(param_group["params"]) > 0:
                        optimizer_device = param_group["params"][0].device
                        break
        return (model_device, optimizer_device)
    def get_state_dict(self, model, unwrap=True):
        if self.distributed_type == DistributedType.DEEPSPEED:
            if self.deepspeed_config["zero_optimization"]["stage"] == 3:
                if model.zero_gather_16bit_weights_on_model_save(): state_dict = model._zero3_consolidated_16bit_state_dict()
                else: raise ValueError("Cannot get 16bit model weights because `stage3_gather_16bit_weights_on_model_save` in DeepSpeed config is False. To save the model weights in 16bit, set `stage3_gather_16bit_weights_on_model_save` to True in DeepSpeed config file or set `zero3_save_16bit_model` to True when using `sapiens_accelerator config`. To save the full checkpoint, run `model.save_checkpoint(save_dir)` and use `zero_to_fp32.py` to recover weights.")
            else:
                from deepspeed.checkpoint.utils import clone_tensors_for_torch_save
                state_dict = clone_tensors_for_torch_save(self.unwrap_model(model).state_dict())
        elif self.distributed_type == DistributedType.FSDP:
            from torch.distributed.fsdp import FullStateDictConfig, StateDictType
            from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
            full_state_dict_config = FullStateDictConfig(offload_to_cpu=True, rank0_only=True)
            with FSDP.state_dict_type(model, StateDictType.FULL_STATE_DICT, full_state_dict_config): state_dict = model.state_dict()
        else:
            if unwrap: model = self.unwrap_model(model)
            state_dict = model.state_dict()
        return state_dict
    def register_for_checkpointing(self, *objects):
        invalid_objects = []
        for obj in objects:
            if not hasattr(obj, "state_dict") or not hasattr(obj, "load_state_dict"): invalid_objects.append(obj)
        if len(invalid_objects) > 0:
            err = "All `objects` must include a `state_dict` and `load_state_dict` function to be stored. The following inputs are invalid:"
            for index, obj in enumerate(invalid_objects): err += f"\n\t- Item at index {index}, `{get_pretty_name(obj)}`"
            raise ValueError(err)
        self._custom_objects.extend(objects)
    @contextmanager
    def autocast(self, autocast_handler: AutocastKwargs = None):
        if autocast_handler is None: autocast_handler = self.autocast_handler
        autocast_context = get_mixed_precision_context_manager(self.native_amp, autocast_handler)
        autocast_context.__enter__()
        yield
        autocast_context.__exit__(*sys.exc_info())
    @contextmanager
    def profile(self, profile_handler: ProfileKwargs | None = None):
        profile_handler = profile_handler or self.profile_handler or ProfileKwargs()
        with profile_handler.build() as profiler: yield profiler
        if profile_handler.output_trace_dir is None: return
        os.makedirs(profile_handler.output_trace_dir, exist_ok=True)
        profiler.export_chrome_trace(os.path.join(profile_handler.output_trace_dir, PROFILE_PATTERN_NAME.format(suffix=self.process_index)))
        self.wait_for_everyone()
    @property
    def optimizer_step_was_skipped(self):
        for optimizer in self._optimizers:
            if optimizer.step_was_skipped: return True
        return False
    def skip_first_batches(self, dataloader, num_batches: int = 0): return skip_first_batches(dataloader, num_batches=num_batches)
    def __deepcopy__(self, memo): return self
    def verify_device_map(self, model: torch.nn.Module) -> bool:
        for m in model.modules():
            if hasattr(m, "hf_device_map") and len(m.hf_device_map) > 1: return True
        return False
    def lomo_backward(self, loss: torch.Tensor, learning_rate: float) -> None:
        if is_lomo_available(): from lomo_optim import AdaLomo, Lomo
        if learning_rate is None: raise ValueError("A learning rate must be passed in order to call backward pass with LOMO optimizers.")
        _backward_called = False
        for optimizer in self._optimizers:
            if isinstance(optimizer.optimizer, (Lomo, AdaLomo)):
                optimizer.optimizer.fused_backward(loss, learning_rate)
                _backward_called = True
        if not _backward_called: raise ValueError("Backward pass not properly called on LOMO optimizers. Are you sure you passed a LOMO optimizer in accelerator.prepare()?")
    @property
    def fp8_backend(self):
        if self.mixed_precision == "fp8" and self.fp8_recipe_handler is not None: return self.fp8_recipe_handler.backend
        elif self.state.deepspeed_plugin is not None and self.state.deepspeed_plugin.enable_msamp: return "MSAMP"
        return None
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import functools
import gc
import importlib
import inspect
import warnings
import torch
from .imports import (is_cuda_available, is_ipex_available, is_mlu_available, is_mps_available, is_musa_available, is_npu_available, is_xpu_available)
from .versions import compare_versions
def clear_device_cache(garbage_collection=False):
    if garbage_collection: gc.collect()
    if is_xpu_available(): torch.xpu.empty_cache()
    elif is_mlu_available(): torch.mlu.empty_cache()
    elif is_musa_available(): torch.musa.empty_cache()
    elif is_npu_available(): torch.npu.empty_cache()
    elif is_mps_available(min_version="2.0"): torch.mps.empty_cache()
    elif is_cuda_available(): torch.cuda.empty_cache()
def release_memory(*objects):
    if not isinstance(objects, list): objects = list(objects)
    for i in range(len(objects)): objects[i] = None
    clear_device_cache(garbage_collection=True)
    return objects
def should_reduce_batch_size(exception: Exception) -> bool:
    _statements = ["CUDA out of memory.", "cuDNN error: CUDNN_STATUS_NOT_SUPPORTED.", "DefaultCPUAllocator: can't allocate memory"]
    if isinstance(exception, RuntimeError) and len(exception.args) == 1: return any(err in exception.args[0] for err in _statements)
    return False
def find_executable_batch_size(function: callable = None, starting_batch_size: int = 128):
    if function is None: return functools.partial(find_executable_batch_size, starting_batch_size=starting_batch_size)
    batch_size = starting_batch_size
    def decorator(*args, **kwargs):
        nonlocal batch_size
        clear_device_cache(garbage_collection=True)
        params = list(inspect.signature(function).parameters.keys())
        if len(params) < (len(args) + 1):
            arg_str = ", ".join([f"{arg}={value}" for arg, value in zip(params[1:], args[1:])])
            raise TypeError(f"Batch size was passed into `{function.__name__}` as the first argument when called. Remove this as the decorator already does so: `{function.__name__}({arg_str})`")
        while True:
            if batch_size == 0: raise RuntimeError("No executable batch size found, reached zero.")
            try: return function(batch_size, *args, **kwargs)
            except Exception as e:
                if should_reduce_batch_size(e):
                    clear_device_cache(garbage_collection=True)
                    batch_size //= 2
                else: raise
    return decorator
def get_xpu_available_memory(device_index: int):
    if is_ipex_available():
        ipex_version = importlib.metadata.version("intel_extension_for_pytorch")
        if compare_versions(ipex_version, ">=", "2.5"):
            from intel_extension_for_pytorch.xpu import mem_get_info
            return mem_get_info(device_index)[0]
    return torch.xpu.max_memory_allocated(device_index)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library for optimizing the Artificial Intelligence models of Sapiens Technology®, and its disclosure, distribution, #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from setuptools import setup, find_packages
extras = {}
extras["quality"] = ["black~=23.1", "hf-doc-builder==0.5.0", "ruff~=0.6.4"]
extras["docs"] = []
extras["test_prod"] = ["pytest>=7.2.0,<=8.0.0", "pytest-xdist==3.6.1", "pytest-subtests==0.14.1", "parameterized==0.9.0"]
extras["test_dev"] = ["datasets==3.0.1", "diffusers==0.32.2", "evaluate==0.4.3", "torchdata==0.10.1", "torchpippy==0.2.0", "transformers==4.45.2", "scipy==1.15.1", "scikit-learn==1.5.0", "tqdm==4.66.4", "sapiens-machine==1.0.9", "timm==1.0.14"]
extras["testing"] = extras["test_prod"] + extras["test_dev"]
extras["deepspeed"] = ["deepspeed"]
extras["rich"] = ["rich"]
extras["test_trackers"] = ["wandb==0.19.5", "comet-ml==3.48.1", "tensorboard==2.18.0", "dvclive==3.48.1"]
extras["dev"] = extras["quality"] + extras["testing"] + extras["rich"]
extras["sagemaker"] = ["sagemaker==2.239.0"]
setup(
    name="sapiens_accelerator",
    version="1.0.9",
    license="Proprietary Software",
    author="OPENSAPI",
    url="https://github.com/sapiens-technology/sapiens_accelerator",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "sapiens-accelerator=accelerate.commands.accelerate_cli:main",
            "sapiens-accelerator-config=accelerate.commands.config:main",
            "sapiens-accelerator-estimate-memory=accelerate.commands.estimate:main",
            "sapiens-accelerator-launch=accelerate.commands.launch:main",
            "sapiens-accelerator-merge-weights=accelerate.commands.merge:main"
        ]
    },
    python_requires=">=3.9.0",
    install_requires=["numpy", "packaging", "psutil", "PyYAML", "torch", "huggingface-hub", "safetensors"],
    extras_require=extras
)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library for optimizing the Artificial Intelligence models of Sapiens Technology®, and its disclosure, distribution, #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
from .imports import is_tqdm_available
if is_tqdm_available(): from tqdm.auto import tqdm as _tqdm
from ..state import PartialState
def tqdm(*args, main_process_only: bool = True, **kwargs):
    if not is_tqdm_available(): raise ImportError("SapiensAccelerator's `tqdm` module requires `tqdm` to be installed. Please run `pip install tqdm`.")
    if len(args) > 0 and isinstance(args[0], bool): raise ValueError("Passing `True` or `False` as the first argument to SapiensAccelerator's `tqdm` wrapper is unsupported. Please use the `main_process_only` keyword argument instead.")
    disable = kwargs.pop("disable", False)
    if main_process_only and not disable: disable = PartialState().local_process_index != 0
    return _tqdm(*args, **kwargs, disable=disable)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import functools
import logging
import os
from .state import PartialState
class MultiProcessAdapter(logging.LoggerAdapter):
    @staticmethod
    def _should_log(main_process_only):
        state = PartialState()
        return not main_process_only or (main_process_only and state.is_main_process)
    def log(self, level, msg, *args, **kwargs):
        if PartialState._shared_state == {}: raise RuntimeError("You must initialize the sapiens_accelerator state by calling either `PartialState()` or `Accelerator()` before using the logging utility.")
        main_process_only = kwargs.pop("main_process_only", True)
        in_order = kwargs.pop("in_order", False)
        kwargs.setdefault("stacklevel", 2)
        if self.isEnabledFor(level):
            if self._should_log(main_process_only):
                msg, kwargs = self.process(msg, kwargs)
                self.logger.log(level, msg, *args, **kwargs)
            elif in_order:
                state = PartialState()
                for i in range(state.num_processes):
                    if i == state.process_index:
                        msg, kwargs = self.process(msg, kwargs)
                        self.logger.log(level, msg, *args, **kwargs)
                    state.wait_for_everyone()
    @functools.lru_cache(None)
    def warning_once(self, *args, **kwargs): self.warning(*args, **kwargs)
def get_logger(name: str, log_level: str = None):
    if log_level is None: log_level = os.environ.get("SAPIENS_ACCELERATOR_LOG_LEVEL", None)
    logger = logging.getLogger(name)
    if log_level is not None:
        logger.setLevel(log_level.upper())
        logger.root.setLevel(log_level.upper())
    return MultiProcessAdapter(logger, {})
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import logging
import math
import os
import platform
import subprocess
import sys
from dataclasses import dataclass, field
from functools import lru_cache
from shutil import which
from typing import List, Optional
import torch
from packaging.version import parse
logger = logging.getLogger(__name__)
def convert_dict_to_env_variables(current_env: dict):
    forbidden_chars = [";", "\n", "<", ">", " "]
    valid_env_items = []
    for key, value in current_env.items():
        if all(char not in (key + value) for char in forbidden_chars) and len(key) >= 1 and len(value) >= 1: valid_env_items.append(f"{key}={value}\n")
    return valid_env_items
def str_to_bool(value) -> int:
    value = value.lower()
    if value in ("y", "yes", "t", "true", "on", "1"): return 1
    elif value in ("n", "no", "f", "false", "off", "0"): return 0
    else: raise ValueError(f"invalid truth value {value}")
def get_int_from_env(env_keys, default):
    for e in env_keys:
        val = int(os.environ.get(e, -1))
        if val >= 0: return val
    return default
def parse_flag_from_env(key, default=False):
    value = os.environ.get(key, str(default))
    return str_to_bool(value) == 1
def parse_choice_from_env(key, default="no"):
    value = os.environ.get(key, str(default))
    return value
def are_libraries_initialized(*library_names: str) -> List[str]: return [lib_name for lib_name in library_names if lib_name in sys.modules.keys()]
def _nvidia_smi():
    if platform.system() == "Windows":
        command = which("nvidia-smi")
        if command is None: command = f"{os.environ['systemdrive']}\\Program Files\\NVIDIA Corporation\\NVSMI\\nvidia-smi.exe"
    else: command = "nvidia-smi"
    return command
def get_gpu_info():
    output = subprocess.check_output([_nvidia_smi(), "--query-gpu=count,name", "--format=csv,noheader"], universal_newlines=True)
    output = output.strip()
    gpus = output.split(os.linesep)
    gpu_count = len(gpus)
    gpu_names = [gpu.split(",")[1].strip() for gpu in gpus]
    return gpu_names, gpu_count
def get_driver_version():
    output = subprocess.check_output([_nvidia_smi(), "--query-gpu=driver_version", "--format=csv,noheader"], universal_newlines=True)
    output = output.strip()
    return output.split(os.linesep)[0]
def check_cuda_p2p_ib_support():
    try:
        device_names, device_count = get_gpu_info()
        unsupported_devices = {"RTX 40"}
        if device_count > 1:
            if any(unsupported_device in device_name for device_name in device_names for unsupported_device in unsupported_devices):
                acceptable_driver_version = "550.40.07"
                current_driver_version = get_driver_version()
                if parse(current_driver_version) < parse(acceptable_driver_version): return False
                return True
    except Exception: pass
    return True
def check_fp8_capability():
    cuda_device_capacity = torch.cuda.get_device_capability()
    return cuda_device_capacity >= (8, 9)
@dataclass
class CPUInformation:
    rank: int = field(default=0, metadata={"help": "The rank of the current process."})
    world_size: int = field(default=1, metadata={"help": "The total number of processes in the world."})
    local_rank: int = field(default=0, metadata={"help": "The rank of the current process on the local node."})
    local_world_size: int = field(default=1, metadata={"help": "The total number of processes on the local node."})
def get_cpu_distributed_information() -> CPUInformation:
    information = {}
    information["rank"] = get_int_from_env(["RANK", "PMI_RANK", "OMPI_COMM_WORLD_RANK", "MV2_COMM_WORLD_RANK"], 0)
    information["world_size"] = get_int_from_env(["WORLD_SIZE", "PMI_SIZE", "OMPI_COMM_WORLD_SIZE", "MV2_COMM_WORLD_SIZE"], 1)
    information["local_rank"] = get_int_from_env(["LOCAL_RANK", "MPI_LOCALRANKID", "OMPI_COMM_WORLD_LOCAL_RANK", "MV2_COMM_WORLD_LOCAL_RANK"], 0)
    information["local_world_size"] = get_int_from_env(["LOCAL_WORLD_SIZE", "MPI_LOCALNRANKS", "OMPI_COMM_WORLD_LOCAL_SIZE", "MV2_COMM_WORLD_LOCAL_SIZE"], 1)
    return CPUInformation(**information)
def override_numa_affinity(local_process_index: int, verbose: Optional[bool] = None) -> None:
    if verbose is None: verbose = parse_flag_from_env("SAPIENS_ACCELERATOR_DEBUG_MODE", False)
    if torch.cuda.is_available():
        from sapiens_accelerator.utils import is_pynvml_available
        if not is_pynvml_available(): raise ImportError("To set CPU affinity on CUDA GPUs the `pynvml` package must be available. (`pip install pynvml`)")
        import pynvml as nvml
        nvml.nvmlInit()
        num_elements = math.ceil(os.cpu_count() / 64)
        handle = nvml.nvmlDeviceGetHandleByIndex(local_process_index)
        affinity_string = ""
        for j in nvml.nvmlDeviceGetCpuAffinity(handle, num_elements): affinity_string = f"{j:064b}{affinity_string}"
        affinity_list = [int(x) for x in affinity_string]
        affinity_list.reverse()
        affinity_to_set = [i for i, e in enumerate(affinity_list) if e != 0]
        os.sched_setaffinity(0, affinity_to_set)
        if verbose: cpu_cores = os.sched_getaffinity(0)
@lru_cache
def set_numa_affinity(local_process_index: int, verbose: Optional[bool] = None) -> None: override_numa_affinity(local_process_index=local_process_index, verbose=verbose)
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import pickle
import warnings
from contextlib import contextmanager, nullcontext
from functools import update_wrapper, wraps
from typing import Any, Mapping
import torch
from ..state import AcceleratorState, PartialState
from .constants import TORCH_DISTRIBUTED_OPERATION_TYPES
from .dataclasses import DistributedType, TensorInformation
from .imports import (is_npu_available, is_torch_distributed_available, is_torch_version, is_torch_xla_available, is_xpu_available)
if is_torch_xla_available(): import torch_xla.core.xla_model as xm
if is_torch_distributed_available(): from torch.distributed import ReduceOp
def is_torch_tensor(tensor): return isinstance(tensor, torch.Tensor)
def is_torch_xpu_tensor(tensor): return isinstance(tensor, (torch.xpu.FloatTensor, torch.xpu.ByteTensor, torch.xpu.IntTensor, torch.xpu.LongTensor, torch.xpu.HalfTensor, torch.xpu.DoubleTensor, torch.xpu.BFloat16Tensor))
def is_tensor_information(tensor_info): return isinstance(tensor_info, TensorInformation)
def is_namedtuple(data): return isinstance(data, tuple) and hasattr(data, "_asdict") and hasattr(data, "_fields")
def honor_type(obj, generator):
    if is_namedtuple(obj): return type(obj)(*list(generator))
    else: return type(obj)(generator)
def recursively_apply(func, data, *args, test_type=is_torch_tensor, error_on_other_type=False, **kwargs):
    if isinstance(data, (tuple, list)): return honor_type(data, (recursively_apply(func, o, *args, test_type=test_type, error_on_other_type=error_on_other_type, **kwargs) for o in data))
    elif isinstance(data, Mapping): return type(data)({k: recursively_apply(func, v, *args, test_type=test_type, error_on_other_type=error_on_other_type, **kwargs) for k, v in data.items()})
    elif test_type(data): return func(data, *args, **kwargs)
    elif error_on_other_type: raise TypeError(f"Unsupported types ({type(data)}) passed to `{func.__name__}`. Only nested list/tuple/dicts of objects that are valid for `{test_type.__name__}` should be passed.")
    return data
def send_to_device(tensor, device, non_blocking=False, skip_keys=None):
    if is_torch_tensor(tensor) or hasattr(tensor, "to"):
        if device == "npu": device = "npu:0"
        if device == "xpu": device = "xpu:0"
        try: return tensor.to(device, non_blocking=non_blocking)
        except TypeError: return tensor.to(device)
        except AssertionError as error:
            if is_npu_available():
                if isinstance(device, int): device = f"npu:{device}"
            elif is_xpu_available():
                if isinstance(device, int): device = f"xpu:{device}"
            else: raise error
        try: return tensor.to(device, non_blocking=non_blocking)
        except TypeError: return tensor.to(device)
    elif isinstance(tensor, (tuple, list)): return honor_type(tensor, (send_to_device(t, device, non_blocking=non_blocking, skip_keys=skip_keys) for t in tensor))
    elif isinstance(tensor, Mapping):
        if isinstance(skip_keys, str): skip_keys = [skip_keys]
        elif skip_keys is None: skip_keys = []
        return type(tensor)({k: t if k in skip_keys else send_to_device(t, device, non_blocking=non_blocking, skip_keys=skip_keys) for k, t in tensor.items()})
    else: return tensor
def get_data_structure(data):
    def _get_data_structure(tensor): return TensorInformation(shape=tensor.shape, dtype=tensor.dtype)
    return recursively_apply(_get_data_structure, data)
def get_shape(data):
    def _get_shape(tensor): return list(tensor.shape)
    return recursively_apply(_get_shape, data)
def initialize_tensors(data_structure):
    def _initialize_tensor(tensor_info): return torch.empty(*tensor_info.shape, dtype=tensor_info.dtype)
    return recursively_apply(_initialize_tensor, data_structure, test_type=is_tensor_information)
def find_batch_size(data):
    if isinstance(data, (tuple, list, Mapping)) and (len(data) == 0): raise ValueError(f"Cannot find the batch size from empty {type(data)}.")
    if isinstance(data, (tuple, list)): return find_batch_size(data[0])
    elif isinstance(data, Mapping):
        for k in data.keys(): return find_batch_size(data[k])
    elif not isinstance(data, torch.Tensor): raise TypeError(f"Can only find the batch size of tensors but got {type(data)}.")
    return data.shape[0]
def ignorant_find_batch_size(data):
    try: return find_batch_size(data)
    except (ValueError, TypeError): pass
    return None
def listify(data):
    def _convert_to_list(tensor):
        tensor = tensor.detach().cpu()
        if tensor.dtype == torch.bfloat16: tensor = tensor.to(torch.float32)
        return tensor.tolist()
    return recursively_apply(_convert_to_list, data)
def _tpu_gather(tensor):
    def _tpu_gather_one(tensor):
        if tensor.ndim == 0: tensor = tensor.clone()[None]
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        return xm.all_gather(tensor)
    res = recursively_apply(_tpu_gather_one, tensor, error_on_other_type=True)
    xm.mark_step()
    return res
def _gpu_gather(tensor):
    state = PartialState()
    if is_torch_version(">=", "1.13"): gather_op = torch.distributed.all_gather_into_tensor
    else: gather_op = torch.distributed._all_gather_base
    def _gpu_gather_one(tensor):
        if tensor.ndim == 0: tensor = tensor.clone()[None]
        if not tensor.is_contiguous(): tensor = tensor.contiguous()
        if state.backend is not None and state.backend != "gloo":
            output_tensors = torch.empty(state.num_processes * tensor.numel(), dtype=tensor.dtype, device=state.device)
            gather_op(output_tensors, tensor)
            return output_tensors.view(-1, *tensor.size()[1:])
        else:
            output_tensors = [torch.empty_like(tensor) for _ in range(state.num_processes)]
            torch.distributed.all_gather(output_tensors, tensor)
            return torch.cat(output_tensors, dim=0)
    return recursively_apply(_gpu_gather_one, tensor, error_on_other_type=True)
class DistributedOperationException(Exception): pass
def verify_operation(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        if PartialState().distributed_type == DistributedType.NO or not PartialState().debug: return function(*args, **kwargs)
        operation = f"{function.__module__}.{function.__name__}"
        if "tensor" in kwargs: tensor = kwargs["tensor"]
        else: tensor = args[0]
        if PartialState().device.type != find_device(tensor).type: raise DistributedOperationException(f"One or more of the tensors passed to {operation} were not on the {tensor.device.type} while the `Accelerator` is configured for {PartialState().device.type}. Please move it to the {PartialState().device.type} before calling {operation}.")
        shapes = get_shape(tensor)
        output = gather_object([shapes])
        if output[0] is not None:
            are_same = output.count(output[0]) == len(output)
            if not are_same:
                process_shape_str = "\n  - ".join([f"Process {i}: {shape}" for i, shape in enumerate(output)])
                raise DistributedOperationException(f"Cannot apply desired operation due to shape mismatches. All shapes across devices must be valid.\n\nOperation: `{operation}`\nInput shapes:\n  - {process_shape_str}")
        return function(*args, **kwargs)
    return wrapper
def chained_operation(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        try: return function(*args, **kwargs)
        except DistributedOperationException as e:
            operation = f"{function.__module__}.{function.__name__}"
            raise DistributedOperationException(f"Error found while calling `{operation}`. Please see the earlier error for more details.") from e
    return wrapper
@verify_operation
def gather(tensor):
    if PartialState().distributed_type == DistributedType.XLA: return _tpu_gather(tensor)
    elif PartialState().distributed_type in TORCH_DISTRIBUTED_OPERATION_TYPES: return _gpu_gather(tensor)
    else: return tensor
def _gpu_gather_object(object: Any):
    output_objects = [None for _ in range(PartialState().num_processes)]
    torch.distributed.all_gather_object(output_objects, object)
    return [x for y in output_objects for x in y]
def gather_object(object: Any):
    if PartialState().distributed_type == DistributedType.XLA: raise NotImplementedError("gather objects in TPU is not supported")
    elif PartialState().distributed_type in TORCH_DISTRIBUTED_OPERATION_TYPES: return _gpu_gather_object(object)
    else: return object
def _gpu_broadcast(data, src=0):
    def _gpu_broadcast_one(tensor, src=0):
        torch.distributed.broadcast(tensor, src=src)
        return tensor
    return recursively_apply(_gpu_broadcast_one, data, error_on_other_type=True, src=src)
def _tpu_broadcast(tensor, src=0, name="broadcast tensor"):
    if isinstance(tensor, (list, tuple)): return honor_type(tensor, (_tpu_broadcast(t, name=f"{name}_{i}") for i, t in enumerate(tensor)))
    elif isinstance(tensor, Mapping): return type(tensor)({k: _tpu_broadcast(v, name=f"{name}_{k}") for k, v in tensor.items()})
    return xm.mesh_reduce(name, tensor, lambda x: x[src])
TENSOR_TYPE_TO_INT = {torch.float: 1, torch.double: 2, torch.half: 3, torch.bfloat16: 4, torch.uint8: 5, torch.int8: 6, torch.int16: 7, torch.int32: 8, torch.int64: 9, torch.bool: 10}
TENSOR_INT_TO_DTYPE = {v: k for k, v in TENSOR_TYPE_TO_INT.items()}
def gather_tensor_shape(tensor):
    max_tensor_dimension = 2**20
    state = PartialState()
    base_tensor = torch.empty(max_tensor_dimension, dtype=torch.int, device=state.device)
    if tensor is not None:
        shape = tensor.shape
        tensor_dtype = TENSOR_TYPE_TO_INT[tensor.dtype]
        base_tensor[: len(shape) + 1] = torch.tensor(list(shape) + [tensor_dtype], dtype=int)
    base_tensor = reduce(base_tensor, reduction="sum")
    base_tensor = base_tensor[base_tensor.nonzero()]
    dtype = int(base_tensor[-1:][0])
    base_tensor = base_tensor[:-1]
    return base_tensor, dtype
def copy_tensor_to_devices(tensor=None) -> torch.Tensor:
    state = PartialState()
    shape, dtype = gather_tensor_shape(tensor)
    if tensor is None: tensor = torch.zeros(shape, dtype=TENSOR_INT_TO_DTYPE[dtype]).to(state.device)
    return reduce(tensor, reduction="sum")
@verify_operation
def broadcast(tensor, from_process: int = 0):
    if PartialState().distributed_type == DistributedType.XLA: return _tpu_broadcast(tensor, src=from_process, name="sapiens_accelerator.utils.broadcast")
    elif PartialState().distributed_type in TORCH_DISTRIBUTED_OPERATION_TYPES: return _gpu_broadcast(tensor, src=from_process)
    else: return tensor
def broadcast_object_list(object_list, from_process: int = 0):
    if PartialState().distributed_type == DistributedType.XLA:
        for i, obj in enumerate(object_list): object_list[i] = xm.mesh_reduce("sapiens_accelerator.utils.broadcast_object_list", obj, lambda x: x[from_process])
    elif PartialState().distributed_type in TORCH_DISTRIBUTED_OPERATION_TYPES: torch.distributed.broadcast_object_list(object_list, src=from_process)
    return object_list
def slice_tensors(data, tensor_slice, process_index=None, num_processes=None):
    def _slice_tensor(tensor, tensor_slice): return tensor[tensor_slice]
    return recursively_apply(_slice_tensor, data, tensor_slice)
def concatenate(data, dim=0):
    if isinstance(data[0], (tuple, list)): return honor_type(data[0], (concatenate([d[i] for d in data], dim=dim) for i in range(len(data[0]))))
    elif isinstance(data[0], Mapping): return type(data[0])({k: concatenate([d[k] for d in data], dim=dim) for k in data[0].keys()})
    elif not isinstance(data[0], torch.Tensor): raise TypeError(f"Can only concatenate tensors but got {type(data[0])}")
    return torch.cat(data, dim=dim)
class CannotPadNestedTensorWarning(UserWarning): pass
@chained_operation
def pad_across_processes(tensor, dim=0, pad_index=0, pad_first=False):
    def _pad_across_processes(tensor, dim=0, pad_index=0, pad_first=False):
        if getattr(tensor, "is_nested", False): return tensor
        if dim >= len(tensor.shape): return tensor
        size = torch.tensor(tensor.shape, device=tensor.device)[None]
        sizes = gather(size).cpu()
        max_size = max(s[dim] for s in sizes)
        if max_size == tensor.shape[dim]: return tensor
        old_size = tensor.shape
        new_size = list(old_size)
        new_size[dim] = max_size
        new_tensor = tensor.new_zeros(tuple(new_size)) + pad_index
        if pad_first: indices = tuple(slice(max_size - old_size[dim], max_size) if i == dim else slice(None) for i in range(len(new_size)))
        else: indices = tuple(slice(0, old_size[dim]) if i == dim else slice(None) for i in range(len(new_size)))
        new_tensor[indices] = tensor
        return new_tensor
    return recursively_apply(_pad_across_processes, tensor, error_on_other_type=True, dim=dim, pad_index=pad_index, pad_first=pad_first)
def pad_input_tensors(tensor, batch_size, num_processes, dim=0):
    def _pad_input_tensors(tensor, batch_size, num_processes, dim=0):
        remainder = batch_size // num_processes
        last_inputs = batch_size - (remainder * num_processes)
        if batch_size // num_processes == 0: to_pad = num_processes - batch_size
        else: to_pad = num_processes - (batch_size // num_processes)
        if last_inputs > to_pad & to_pad < 1: to_pad = last_inputs - to_pad
        old_size = tensor.shape
        new_size = list(old_size)
        new_size[0] = batch_size + to_pad
        new_tensor = tensor.new_zeros(tuple(new_size))
        indices = tuple(slice(0, old_size[dim]) if i == dim else slice(None) for i in range(len(new_size)))
        new_tensor[indices] = tensor
        return new_tensor
    return recursively_apply(_pad_input_tensors, tensor, error_on_other_type=True, batch_size=batch_size, num_processes=num_processes, dim=dim)
@verify_operation
def reduce(tensor, reduction="mean", scale=1.0):
    def _reduce_across_processes(tensor, reduction="mean", scale=1.0):
        state = PartialState()
        cloned_tensor = tensor.clone()
        if state.distributed_type == DistributedType.NO: return cloned_tensor
        if state.distributed_type == DistributedType.XLA:
            xm.mark_step()
            xm.all_reduce(xm.REDUCE_SUM, [cloned_tensor], scale)
            xm.mark_step()
        elif state.distributed_type.value in TORCH_DISTRIBUTED_OPERATION_TYPES: torch.distributed.all_reduce(cloned_tensor, ReduceOp.SUM)
        if reduction == "mean": cloned_tensor /= state.num_processes
        return cloned_tensor
    return recursively_apply(_reduce_across_processes, tensor, error_on_other_type=True, reduction=reduction, scale=scale)
def convert_to_fp32(tensor):
    def _convert_to_fp32(tensor): return tensor.float()
    def _is_fp16_bf16_tensor(tensor): return (is_torch_tensor(tensor) or hasattr(tensor, "dtype")) and tensor.dtype in (torch.float16, torch.bfloat16)
    return recursively_apply(_convert_to_fp32, tensor, test_type=_is_fp16_bf16_tensor)
class ConvertOutputsToFp32:
    def __init__(self, model_forward):
        self.model_forward = model_forward
        update_wrapper(self, model_forward)
    def __call__(self, *args, **kwargs): return convert_to_fp32(self.model_forward(*args, **kwargs))
    def __getstate__(self): raise pickle.PicklingError("Cannot pickle a prepared model with automatic mixed precision, please unwrap the model with `Accelerator.unwrap_model(model)` before pickling it.")
def convert_outputs_to_fp32(model_forward):
    model_forward = ConvertOutputsToFp32(model_forward)
    def forward(*args, **kwargs): return model_forward(*args, **kwargs)
    forward.__wrapped__ = model_forward
    return forward
def find_device(data):
    if isinstance(data, Mapping):
        for obj in data.values():
            device = find_device(obj)
            if device is not None: return device
    elif isinstance(data, (tuple, list)):
        for obj in data:
            device = find_device(obj)
            if device is not None: return device
    elif isinstance(data, torch.Tensor): return data.device
@contextmanager
def GatheredParameters(params, modifier_rank=None, fwd_module=None, enabled=True):
    if AcceleratorState().distributed_type != DistributedType.DEEPSPEED or (AcceleratorState().deepspeed_plugin is not None and not AcceleratorState().deepspeed_plugin.is_zero3_init_enabled()): gather_param_context = nullcontext()
    else:
        import deepspeed
        gather_param_context = deepspeed.zero.GatheredParameters(params, modifier_rank=modifier_rank, fwd_module=fwd_module, enabled=enabled)
    with gather_param_context: yield
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""



"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
import torch.distributed
from sapiens_accelerator.test_utils import require_huggingface_suite, torch_device
from sapiens_accelerator.utils import is_transformers_available
if is_transformers_available(): from transformers import AutoModel, TrainingArguments
GPT2_TINY = "sshleifer/tiny-gpt2"
@require_huggingface_suite
def init_torch_dist_then_launch_deepspeed():
    backend = "ccl" if torch_device == "xpu" else "nccl"
    torch.distributed.init_process_group(backend=backend)
    deepspeed_config = {"zero_optimization": {"stage": 3}, "train_batch_size": "auto", "train_micro_batch_size_per_gpu": "auto"}
    train_args = TrainingArguments(output_dir="./", deepspeed=deepspeed_config)
    model = AutoModel.from_pretrained(GPT2_TINY)
    assert train_args is not None
    assert model is not None
def main(): init_torch_dist_then_launch_deepspeed()
if __name__ == "__main__": main()
"""
    ########################################################################################################################################################
    # This algorithm is part of a code library to accelerate Sapiens Technology® Artificial Intelligence models, and its disclosure, distribution,         #
    # or reverse engineering without the company's prior consent will be subject to legal proceedings and actions pursued by our legal department.         #
    ########################################################################################################################################################
"""
